###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Download singularity image"""
from os.path import exists
from easydev import md5


__all__ = ["download_singularity_image"]


def download_singularity_image(outfile, container_path, md5value=None, force=False):

    assert outfile.endswith(".simg"), "output filename must be .simg"

    # download singularity
    from bioconvert import configuration as config

    # note that in singularity v2.4, whatever extension you put, it is
    # replaced by simg
    singfile = "{}/{}".format(config.user_config_dir, outfile)

    if exists(singfile) and md5value and md5(singfile) == md5value and force is False:
        print("Found singularity (graphviz) image")
    elif exists(singfile) and md5value is None and force is False:
        print("Found singularity (graphviz) image but md5 not checked")
    else:
        print("Downloading singularity. Please wait")
        cmd = "singularity pull --name {}  {}"
        if force is True:
            cmd += " -F "
        cmd = cmd.format(singfile, container_path)
        print(cmd)
        try:
            from bioconvert.core.shell import shell

            shell(cmd)
        except:
            raise Exception
    return singfile



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`BZ2` to :term:`GZ` format"""
import bz2
import gzip

from bioconvert import ConvBase
from bioconvert.core.decorators import requires, requires_nothing

import colorlog

logger = colorlog.getLogger(__name__)


__all__ = ["BZ22GZ"]


class BZ22GZ(ConvBase):
    """Convert :term:`BZ2` file to :term:`GZ` file

    Methods based on bunzip2 or zlib/bz2 Python libraries.

    """

    #: Default value
    _default_method = "bz2_gz"
    _threading = True

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input BZ2 file
        :param str outfile: output GZ filename

        """
        super(BZ22GZ, self).__init__(infile, outfile, *args, **kargs)

    @requires("bunzip2")
    def _method_bz2_gz(self, *args, **kwargs):
        """Method that uses bunzip2 gzip.
        
        `bunzip2 documentation <https://docs.oracle.com/cd/E86824_01/html/E54763/bunzip2-1.html>`_
        `gzip documentation <https://www.gnu.org/software/gzip/manual/gzip.html>`_"""
        # conversion
        cmd = "bunzip2 -c {input} | gzip > {output}".format(
            input=self.infile, output=self.outfile
        )
        self.execute(cmd)

    @requires_nothing
    def _method_python(self, *args, **kargs):
        """Internal method"""
        with bz2.open(self.infile, "rb") as f, gzip.open(self.outfile, "wb") as g:
            g.write(f.read())



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`FASTQ` to :term:`QUAL` format"""
from bioconvert import ConvBase, bioconvert_script
from bioconvert.core.base import ConvArg
from bioconvert.core.decorators import (
    compressor,
    out_compressor,
    in_gz,
    requires,
    requires_nothing,
)
from bioconvert import logger

logger.__name__ = "fastq2qual"


class FASTQ2QUAL(ConvBase):
    """Convert :term:`FASTQ` to :term:`QUAL`"""

    # use readfq for now because pure python are fast enough
    # for production, could use seqtk which seems the fastest method though
    # Make sure that the default handles also the compresssion
    #: Default value
    _default_method = "readfq"

    # (https://raw.githubusercontent.com/lh3/readfq/master/readfq.py)
    @staticmethod
    def _readfq(fp):  # this is a generator function
        last = None  # this is a buffer keeping the last unprocessed line
        while True:  # mimic closure; is it a bad idea?
            if not last:  # the first record or a record following a fastq
                for l in fp:  # search for the start of the next record
                    if l[0] == "@":  # fastq header line
                        last = l[:-1]  # save this line
                        break
            if not last:
                break
            header, seqs, last = last[1:], [], None
            for l in fp:  # read the sequence
                if l[0] in "@+":
                    last = l[:-1]
                    break
                seqs.append(l[:-1])
            seq, leng, seqs = "".join(seqs), 0, []
            for l in fp:  # read the quality
                seqs.append(l[:-1])
                leng += len(l) - 1
                if leng >= len(seq):  # have read enough quality
                    last = None
                    yield header, seq, "".join(seqs)  # yield a fastq record
                    break

    def __init__(self, infile, outfile):
        """
        :param str infile: The path to the input FASTA file.
        :param str outfile: The path to the output file.
        """
        super(FASTQ2QUAL, self).__init__(infile, outfile)

    @requires_nothing
    @compressor
    def _method_readfq(self, *args, **kwargs):
        """This method is inspired by Readfq coded by Heng Li.
        
        `original Readfq method <https://github.com/lh3/readfq>`_"""
        with open(self.outfile, "w") as outfile, open(self.infile, "r") as fastq:
            for (name, seq, qual) in FASTQ2QUAL._readfq(fastq):
                outfile.write(">{}\n{}\n".format(name, qual))



# -*- coding: utf-8 -*-

###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################

"""
Converter benchmarking
===========================

Converter have a default method.

Notem however, that several methods may be available.
Moreover, you may have a method that you want to compare
with the implemented one. To do so you will need to 
implement your method first. Then, simply use our benchmarking
framework as follows.

"""
#################################################
#
from bioconvert import Benchmark
from bioconvert import bioconvert_data
from bioconvert.bam2cov import BAM2COV
from bioconvert.fastq2fasta import FASTQ2FASTA

#####################################################
# Get the convert you wish to benchmark
input_file = bioconvert_data("test_measles.sorted.bam")
conv = BAM2COV(input_file, "test.cov")
# input_file = bioconvert_data("test_fastq2fasta_v1.fastq")
# conv = FASTQ2FASTA(input_file, "test.fasta")


#####################################################
# Get the Benchmark instance
bench = Benchmark(conv)
bench.plot()

# You can now see the different methods implemented in this
# converter and which one is the fastest.



# -*- coding: utf-8 -*-

###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################


class Fasta:
    """Read a fasta file"""

    def __init__(self, filename):
        self.filename = filename

    def read(self):
        """ Read fasta sequence by sequence creating a generator """
        sequence = {"id": "", "comment": "", "value": ""}

        with open(self.filename) as reader:
            line = None

            for line in reader:
                # Remove line return and useless spaces
                line = " ".join(line.split())

                # Header case
                if line[0] == ">":
                    if len(sequence["id"]) > 0:
                        yield sequence

                    # create new sequence
                    sequence = {"id": "", "comment": "", "value": ""}
                    # Parse header to split into id and comment
                    idx = line.find(" ")
                    if idx == -1:
                        sequence["id"] = line[1:]
                    else:
                        sequence["id"] = line[1:idx]
                        sequence["comment"] = line[idx + 1 :]

                # Sequence case
                else:
                    sequence["value"] = "{}{}".format(sequence["value"], line)

        if len(sequence["id"]) > 0:
            yield sequence



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`GENBANK` to :term:`GFF3` format"""

from bioconvert import ConvBase

__all__ = ["GENBANK2GFF3"]


class GENBANK2GFF3(ConvBase):
    """Convert :term:`GENBANK` file to :term:`GFF3` file

    Method based on biocode.

    """

    #: Default value
    _default_method = "biocode"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input GENBANK file
        :param str outfile: output GFF3 filename

        """
        super(GENBANK2GFF3, self).__init__(infile, outfile)

    def _method_biocode(self, *args, **kwargs):
        """Uses scripts from biocode copied and modified in bioconvert.utils.biocode

        Please see `Main entry  <https://github.com/jorvis/biocode/>`_ 
        """
        from bioconvert.utils.biocode.convert_genbank_to_gff3 import gbk2gff3
        gbk2gff3(self.infile, self.outfile, fasta=False)














#
#  This file is part of Bioconvert software
#
#  Copyright (c) 2017 - Bioconvert Development Team
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/biokit/bioconvert
#  documentation: http://bioconvert.readthedocs.io
#
##############################################################################
"""Convert :term:`BIGWIG` format to :term:`WIGGLE` format"""
from bioconvert import ConvBase
import colorlog

from bioconvert.core.decorators import requires

_log = colorlog.getLogger(__name__)

__all__ = ["BIGWIG2WIGGLE"]


class BIGWIG2WIGGLE(ConvBase):
    """Convert sorted :term:`BIGWIG` file into :term:`WIGGLE` file 

    Methods available are based on pybigwig [DEEPTOOLS]_.

    """

    #: Default value
    _default_method = "wiggletools"

    def __init__(self, infile, outfile):
        """
        :param str infile: The path to the input BIGWIG file. **It must be sorted**.
        :param str outfile: The path to the output file
        """
        super(BIGWIG2WIGGLE, self).__init__(infile, outfile)

    @requires("wiggletools")
    def _method_wiggletools(self, *args, **kwargs):
        """Conversion using wiggletools

        `wiggletools documentation <https://github.com/Ensembl/WiggleTools>`_"""
        cmd = "wiggletools {} > {}".format(self.infile, self.outfile)
        self.execute(cmd)



# -*- coding: utf-8 -*-
#
# documentation build configuration file, created by
# sphinx-quickstart on Wed Aug 10 16:58:13 2011.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys, os
import sphinx

sys.path.insert(0, os.path.abspath("sphinxext"))

import sphinx_gallery

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

pkg_name = "bioconvert"

# This is for ReadTheDoc
import matplotlib

matplotlib.use("Agg")

on_rtd = os.environ.get("READTHEDOCS", None) == "True"

# This creates the conversion.png image automatically

import sys

sys.path.append(".")
# import create_graph


import pkg_resources

version = pkg_resources.require(pkg_name)[0].version

import matplotlib
import matplotlib.sphinxext

release = version
author = "Bioconvet developers"
title = "Bioconvert"
copyright = author + ", 2017-2018"
project = "bioconvert"


import bioconvert

bioconvert.logger.level = "WARNING"

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.

extensions = [
    "sphinx.ext.autodoc",
    (
        "sphinx.ext.imgmath"  # only available for sphinx >= 1.4
        if sphinx.version_info[:2] >= (1, 4)
        else "sphinx.ext.pngmath"
    ),
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "numpydoc.numpydoc",
    "matplotlib.sphinxext.plot_directive",
    "sphinx.ext.autosummary",
    "sphinx_gallery.gen_gallery",
]
# note that the numpy directives is buggy. Example: class and init are not recognised as two entities for the autoclass_content=both here below


todo_include_todos = True
jscopybutton_path = "copybutton.js"
autoclass_content = "both"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = project
copyright = copyright

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = "Current version: " + str(version)
# The full version, including alpha/beta/rc tags.
release = release

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
# unused_docs = []


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_trees = ["build"]
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ["bioconvert."]

# -- sphinx gallery ------------------------------------------------------------

# By default, examples are not built locally. You can set plot_gallery to True
# to force their creation. Note that it requires singularity or dot to be
# installed. Fixes https://github.com/biokit/bioconvert/issues/153

plot_gallery = True

sphinx_gallery_conf = {
    "doc_module": "bioconvert",
    "backreferences_dir": os.path.join("modules", "generated"),
    # "backreferences_dir": False,
    # "filename_pattern": 'plot_benchmark'
}

# Get rid of spurious warnings due to some interaction between
# autosummary and numpydoc. See
# https://github.com/phn/pytpm/issues/3#issuecomment-12133978 for more
# details
numpydoc_show_class_members = False


# solution from nilearn
def touch_example_backreferences(app, what, name, obj, options, lines):
    # generate empty examples files, so that we don't get
    # inclusion errors if there are no examples for a class / module
    examples_path = os.path.join(
        app.srcdir, "modules", "generated", "{}.examples".format(name)
    )
    if not os.path.exists(examples_path):
        # touch file
        os.makedirs(os.path.dirname(examples_path), exist_ok=True)
        open(examples_path, "w").close()


# Add the 'copybutton' javascript, to hide/show the prompt in code
# examples
def setup(app):
    app.add_js_file("copybutton.js")
    app.connect("autodoc-process-docstring", touch_example_backreferences)


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
if not on_rtd:
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
else:
    html_theme = "default"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# the user theme contains the otpions 'homepage', which is populated here

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = [get_path_sphinx_themes()]


# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = "Bioconvert"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".

# the copybutton.js must be copied there:
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%b %d, %Y"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = False

smartquotes = False

# Custom sidebar templates, maps document names to template names.
html_index = "index.html"

# Custom sidebar templates, maps page names to templates.
# html_sidebars = {
#                    'index': [ 'indexsidebar.html'],
#                    'contents':'indexsidebar.html',
# }
# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {   'index': 'index.html'}


# If false, no module index is generated.
html_use_modindex = True
html_domain_indices = True
# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True
html_copy_source = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "doc"


# -- Options for LaTeX output --------------------------------------------------

# NOT in original quickstart
pngmath_use_preview = True

# The font size ('10pt', '11pt' or '12pt').
latex_font_size = "10pt"

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ("index", "main.tex", title, author, "manual"),
]

latex_elements = {"inputenc": "\\usepackage[utf8]{inputenc}"}
# latex_elements["latex_paper_size"] = "a4"
# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False


# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", project, project, [author], 1)]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("http://docs.python.org/", None),
}



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`GZ` to :term:`DSRC` format"""

import colorlog

from bioconvert import ConvBase
from bioconvert import requires

logger = colorlog.getLogger(__name__)


__all__ = ["GZ2DSRC"]


class GZ2DSRC(ConvBase):
    """Convert compressed fastq.gz file into `DSRC` compressed file

    .. plot::

         from bioconvert.gz2dsrc import GZ2DSRC
         from bioconvert import bioconvert_data
         from easydev import TempFile

         with TempFile(suffix=".dsrc") as fh:
             infile = bioconvert_data("test_SP1.fq.gz")
             convert = GZ2DSRC(infile, fh.name)
             convert.boxplot_benchmark()

    """

    #: Default value
    _default_method = "pigzdsrc"
    _threading = True

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input GZ filename
        :param str outfile: output DSRC filename

        """
        super(GZ2DSRC, self).__init__(infile, outfile, *args, **kargs)

    @requires(external_binaries=["pigz", "dsrc"])
    def _method_pigzdsrc(self, *args, **kwargs):
        """do the conversion gz -> :term:`DSRC`

        :return: the standard output
        :rtype: :class:`io.StringIO` object.

        Method that uses pigz and dsrc.

        `pigz documentation <https://linux.die.net/man/1/pigz>`_
        `dsrc documentation <https://github.com/refresh-bio/DSRC>`_"""
        cmd = "pigz -d -c -p {threads} {input} | dsrc c -s -t{threads} {output}"
        self.execute(
            cmd.format(threads=self.threads, input=self.infile, output=self.outfile)
        )



import sys

from bioconvert.utils.biocode import utils
from bioconvert.utils.biocode import things


def go_namespace_index(obo_path):
    """
    Parses a GO OBO file and returns a dict where the keys are GO IDs and values
    are the namespace each belongs to.
    
    Assumes the id field is always defined before the namespace one.
    """
    idx = dict()

    current_id = None
    alts = list()
    current_product = None
    current_namespace = None

    for line in open(obo_path):
        line = line.rstrip()

        if line.startswith('[Term]'):
            idx[current_id] = {'p': current_product, 'n': current_namespace}
            for alt in alts:
                idx[alt] = {'p': current_product, 'n': current_namespace}

            # reset
            current_id = current_product = current_namespace = None
            alts = list()
        elif line.startswith('id:'):
            current_id = line.split()[1]
        elif line.startswith('namespace:'):
            current_namespace = line.split()[1]
        elif line.startswith('name:'):
            current_product = line[6:]
        elif line.startswith('alt_id:'):
            alts.append(line.split()[1])

    # don't forget the last one
    idx[current_id] = {'p': current_product, 'n': current_namespace}
    for alt in alts:
        idx[alt] = {'p': current_product, 'n': current_namespace}

    return idx

def print_tbl_from_assemblies(assemblies=None, ofh=None, go_obo=None, lab_name=None):
    """
    Utility function to write a complete NCBI tbl file from a list() of biothings.Assembly objects

    The go_obo option is required if you want to be able to export GO terms in the annotation.  This
    is because, at the time of this writing, there is no standard in GFF for specifying terms as part of
    the 'process', 'function' or 'component' namespaces but the TBL format requires this distinction
    for each term.  This function indexes it instead so this distinction can be made.

    References:
      Eukaryotic submission examples:
      https://www.ncbi.nlm.nih.gov/genbank/eukaryotic_genome_submission_examples/

    """
    if type(ofh).__name__ != 'TextIOWrapper':
        ofh = sys.stdout #TextIOWrapper

    if go_obo is None:
        go_idx = None
    else:
        go_idx = go_namespace_index(go_obo)
    
    for assembly_id in assemblies:
        current_assembly = assemblies[assembly_id]

        ofh.write(">Feature {0}\n".format(assembly_id))
        
        for gene in sorted(assemblies[assembly_id].genes()):
            rnas_found = 0
            mRNAs = gene.mRNAs()
            
            for mRNA in mRNAs:
                mRNA_loc = mRNA.location_on(current_assembly)
                rnas_found += 1

                if rnas_found > 1:
                    gene.remove_mRNA(mRNA)
                    
                    print("INFO: splitting mRNA off gene {0}".format(gene.id))
                    new_gene = things.Gene(id="{0}_{1}".format(gene.id, rnas_found))
                    
                    new_gene.locate_on(target=current_assembly, fmin=mRNA_loc.fmin, fmax=mRNA_loc.fmax, strand=mRNA_loc.strand)
                    new_gene.add_RNA(mRNA)
                    print_biogene(gene=new_gene, fh=ofh, obo_dict=go_idx, lab_name=lab_name)

            if len(mRNAs) > 1:
                gene_loc = gene.location_on(current_assembly)
                mRNA_loc = mRNAs[0].location_on(current_assembly)
                gene_loc.fmin = mRNA_loc.fmin
                gene_loc.fmax = mRNA_loc.fmax
                gene_loc.strand = mRNA_loc.strand

            print_biogene(gene=gene, fh=ofh, obo_dict=go_idx, lab_name=lab_name)
    
    


def print_biogene( gene=None, fh=None, on=None, obo_dict=None, lab_name=None ):
    '''
    This method accepts a Gene object located on an Assembly object (from things.py) and prints
    the feature graph for that gene in TBL format, including the gene, mRNA, CDS and exon features.

    The obo_dict option is required for export of GO terms.  Keys are GO IDs, like 'GO:0005654' and
    values are one of: 'component', 'process' or 'function'
    '''
    if gene is None:
        raise Exception( "ERROR: The print_biogene() function requires a biogene to be passed via the 'gene' argument" );

    ## we can auto-detect the molecule if the user didn't pass one
    #   and if there's only one.
    if on is None:
        on = gene.location().on

    gene_loc = gene.location_on( on )
    gene_coords = utils.interbase0_to_humancoords(gene_loc.fmin, gene_loc.fmax, gene_loc.strand)

    # Genes MUST have locus tags in TBL format
    gene_annot_atts = dict()
    if gene.locus_tag is None:
        raise Exception("ERROR: locus_tag attributes are required for all gene elements (gene id: {0}".format(gene.id))

    # http://www.ncbi.nlm.nih.gov/genbank/eukaryotic_genome_submission_annotation#protein_id
    official_protein_id = "gnl|{0}|{1}".format(lab_name, gene.locus_tag)
    official_transcript_id = "gnl|{0}|mrna.{1}".format(lab_name, gene.locus_tag)

    fh.write("{0}\t{1}\tgene\n".format(gene_coords[0], gene_coords[1]))
    fh.write("\t\t\tlocus_tag\t{0}\n".format(gene.locus_tag))

    gene_symbol_printed = False

    for RNA in gene.RNAs():
        RNA_loc = RNA.location_on(on)

        if RNA_loc is None:
            raise Exception("ERROR: Expected RNA {0} to be located on {1} but it wasn't".format(RNA.id, on.id))

        # Get polypeptides associated with this.  There should be 1 or 0 (for ncRNAs)
        polypeptides = RNA.polypeptides()

        if len(polypeptides) == 1:
            annot = polypeptides[0].annotation
        elif len(polypeptides) == 0:
            annot = RNA.annotation
        elif len(polypeptides) > 1:
            raise Exception("ERROR: RNAs with multiple polpeptides is currently unsupported: {0}".format(RNA.id))

        if annot is not None and gene_symbol_printed == False:
            if annot.gene_symbol is not None:
                fh.write("\t\t\tgene\t{0}\n".format(annot.gene_symbol))
                gene_symbol_printed = True

        # In this format, the RNA is supposed to be provided as segmented coordinates
        exons_printed = 0
        exons = sorted(RNA.exons())
        if RNA_loc.strand == -1:
            exons.reverse()
            
        for exon in exons:
            exon_loc = exon.location_on(on)
            
            if exon_loc is None:
                raise Exception("ERROR: Expected exon {0} to be located on {1} but it wasn't".format(exon.id, on.id) )

            exon_coords = utils.interbase0_to_humancoords(exon_loc.fmin, exon_loc.fmax, exon_loc.strand)

            if exons_printed == 0:
                fh.write("{0}\t{1}\t{2}\n".format(exon_coords[0], exon_coords[1], RNA.__class__.__name__))
            else:
                fh.write("{0}\t{1}\n".format(exon_coords[0], exon_coords[1]))

            exons_printed += 1

        if annot is not None and exons_printed > 0:
            if RNA.__class__.__name__ == 'mRNA':
                fh.write("\t\t\tprotein_id\t{0}\n".format(official_protein_id))
                
            fh.write("\t\t\ttranscript_id\t{0}\n".format(official_transcript_id))
            fh.write("\t\t\tproduct\t{0}\n".format(annot.product_name))
            if 'Note' in annot.other_attributes:
                # Sometimes this is a string, sometimes a list.  Only strings have the lower function:
                if hasattr(annot.other_attributes['Note'], 'lower'):
                    fh.write("\t\t\tnote\t{0}\n".format(annot.other_attributes['Note']))
                else:
                    fh.write("\t\t\tnote\t{0}\n".format(",".join(annot.other_attributes['Note'])))

        CDS_printed = 0
        CDSs = sorted(RNA.CDSs())
        if RNA_loc.strand == -1:
            CDSs.reverse()
        for CDS in CDSs:
            CDS_loc = CDS.location_on(on)

            if CDS_loc is None:
                raise Exception("ERROR: Expected CDS {0} to be located on {1} but it wasn't".format(CDS.id, on.id) )

            CDS_coords = utils.interbase0_to_humancoords(CDS_loc.fmin, CDS_loc.fmax, CDS_loc.strand)

            if CDS_printed == 0:
                fh.write("{0}\t{1}\tCDS\n".format(CDS_coords[0], CDS_coords[1]))
            else:
                fh.write("{0}\t{1}\n".format(CDS_coords[0], CDS_coords[1]))

            CDS_printed += 1

        if annot is not None and CDS_printed > 0:
            fh.write("\t\t\tprotein_id\t{0}\n".format(official_protein_id))
            fh.write("\t\t\ttranscript_id\t{0}\n".format(official_transcript_id))
            fh.write("\t\t\tproduct\t{0}\n".format(annot.product_name))

            if len(annot.ec_numbers) > 0:
                for ec_number in annot.ec_numbers:
                    fh.write("\t\t\tEC_number\t{0}\n".format(ec_number.number))
        
            if obo_dict is not None and len(annot.go_annotations) > 0:
                for go_term in annot.go_annotations:
                    go_id = "GO:{0}".format(go_term.go_id)
                    
                    """
                    There should actually be a distinction here between ISA (from blast/rapsearch matches)
                    and ISM (from HMM searches) but this is currently not encoded into the GFF, which is
                    our only source so far.  TODO: devise a way to pass this through.  For now, safer
                    to just use the all-encompassing 'ISS'

                    Reference:  http://geneontology.org/page/iss-inferred-sequence-or-structural-similarity
                    """
                    if go_id in obo_dict:
                        if obo_dict[go_id]['n'] == 'molecular_function':
                            fh.write("\t\t\tgo_function\t{1}|{0}||ISS\n".format(go_term.go_id, obo_dict[go_id]['p']))
                        elif obo_dict[go_id]['n'] == 'biological_process':
                            fh.write("\t\t\tgo_process\t{1}|{0}||ISS\n".format(go_term.go_id, obo_dict[go_id]['p']))
                        elif obo_dict[go_id]['n'] == 'cellular_component':
                            fh.write("\t\t\tgo_component\t{1}|{0}||ISS\n".format(go_term.go_id, obo_dict[go_id]['p']))
                    else:
                        raise Exception("ERROR: Didn't find GO id ({0}) in passed OBO file".format(go_id))

            if len(annot.dbxrefs) > 0:
                for dbxref in annot.dbxrefs:
                    fh.write("\t\t\tdb_xref\t{0}:{1}\n".format(dbxref.db, dbxref.identifier))
        

    














##############################################################################
#  This file is part of Bioconvert software
#
#  Copyright (c) 2017 - Bioconvert Development Team
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/biokit/bioconvert
#  documentation: http://bioconvert.readthedocs.io
##############################################################################
"""Convert :term:`ABI` format to :term:`QUAL` format"""
from bioconvert import ConvBase
from bioconvert import requires


__all__ = ["ABI2QUAL"]


class ABI2QUAL(ConvBase):
    """Convert :term:`ABI` file to :term:`QUAL` file

    :term:`ABI` files are created by ABI sequencing machine and
    includes PHRED quality scores for base calls. This allows
    the creation of :term:`QUAL` files.

    Method implemented is based on BioPython [BIOPYTHON]_.

    """

    #: Default value
    _default_method = "biopython"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input ABI file
        :param str outfile: output QUAL filename

        """
        super(ABI2QUAL, self).__init__(infile, outfile, *args, **kargs)

    @requires(python_library="biopython")
    def _method_biopython(self, *args, **kwargs):
        """For this method we use the biopython package Bio.SeqIO.

        `Bio.SeqIO Documentation <https://biopython.org/docs/1.76/api/Bio.SeqIO.html>`_"""
        from Bio import SeqIO

        records = SeqIO.parse(self.infile, "abi")
        # output using SeqIO.write(records, self.outfile, "qual") is not
        # standard so we write our own conversion here below
        with open(self.outfile, "w") as fout:
            for rec in records:
                header = rec.name
                qual = rec.letter_annotations["phred_quality"]
                qual = "".join([str(x) for x in qual])
                fout.write(">{}\n".format(header))
                fout.write("{}\n".format(qual))



'''
Warning: this module requires Python 3.2 or higher

This is a set of classes to represent what I have most commonly needed when working on
dozens (eukaryotic) or hundreds (prokaryotic) of annotation projects.  If you have
other attributes that you'd like to see supported, please add an 'issue' on the
biocode GitHub page.
'''

import re

class FunctionalAnnotation:
    """
    While recognizing that an enormous variety of attributes could go here in
    describing the functional annotation of a BioThing, I'm starting with those
    we most often encounter and need to be available in common output formats.
    These most common attributes are accessed by name, but the others are found
    as a dict in the 'other_attributes' property.

    Also, there's a place for having attributes like this abstracted, stored in
    ontologies, etc.  We've done all that before.  For now I'm going to try
    and hopefully enjoy the utility of having the most common properties always
    directly, and simply, available.
    """
    def __init__( self, product_name=None, gene_symbol=None, go_annotations=None, ec_numbers=None, dbxrefs=None ):
        self.product_name     = product_name
        self.gene_symbol      = gene_symbol
        self.go_annotations   = go_annotations
        self.ec_numbers       = ec_numbers
        self.dbxrefs          = dbxrefs
        self.other_attributes = dict()

        if self.go_annotations is None:
            self.go_annotations = list()

        if self.ec_numbers is None:
            self.ec_numbers = list()

        if self.dbxrefs is None:
            self.dbxrefs = list()

    def __str__(self):
        representation = "Product name: {0}\nGene symbol : {1}\n".format(self.product_name, self.gene_symbol)

        if len(self.go_annotations) > 0:
            representation += "GO annotations:\n"
            for go_annot in self.go_annotations:
                representation += "\tGO:{0}\n".format(go_annot.go_id)
        else:
            representation += "GO annotations: None\n"

        if len(self.ec_numbers) > 0:
            representation += "EC numbers:\n"
            for ec in self.ec_numbers:
                representation += "\t{0}\n".format(ec.number)
        else:
            representation += "EC numbers: None\n"

        if len(self.dbxrefs) > 0:
            representation += "Dbxrefs:\n"
            for dbxref in self.dbxrefs:
                representation += "\t{0}:{1}\n".format(dbxref.db, dbxref.identifier)
        else:
            representation += "Dbxrefs: None\n"

        return representation

    def add_dbxref(self, dbxref):
        """
        Stores a Dbxref object within an annotation. The thing passed can either be a
        Dbxref object or string like "sourcedb:identifier" and it will be automatically
        parsed.
        """
        if type(dbxref).__name__ == 'Dbxref':
            self.dbxrefs.append(dbxref)
        elif type(dbxref).__name__ == 'str':
            m = re.match("(.+)\:(.+)", dbxref)
            if m:
                self.dbxrefs.append(Dbxref(db=m.group(1), identifier=m.group(2)))
            else:
                raise Exception("ERROR: Annotation.add_dbxref(): If string passed, expected format was 'source:identifier'")
        else:
            raise Exception("ERROR: Annotation.add_dbxref expected a Dbxref object or string to be passed")

    def add_ec_number(self, ec_num):
        """
        TODO: Modify this to allow passing ECAnnotation object or string.
        Right now it expects an ECAnnotation object
        """
        self.ec_numbers.append(ec_num)

    def add_go_annotation(self, go):
        """
        TODO: Modify this to allow passing GOAnnotation object or string.
        Right now it expects an GOAnnotation object
        """
        self.go_annotations.append(go)

    def process_gene_symbol(self):
        """
        This method applies a series of rules based on our experience in annotation
        to gene symbols, attempting to correct things before they get submitted to
        GenBank.  This includes:

        - Removing anything after the first whitespace
        - DOES NOT change case.  Genbank demands that prok submissions all use lower
          case but allows organism-specific conventions to be followed for others.
          - https://www.ncbi.nlm.nih.gov/genbank/genomesubmit_annotation/

        This method returns the new gene symbol rather than overwriting the attribute.  For
        that, use set_processed_gene_symbol instead.
        """
        new_gs = self.gene_symbol

        if not new_gs:
            return new_gs

        # take off everything after and including the first space, if any
        new_gs = new_gs.split(' ', 1)[0]

        return new_gs


    def process_product_name(self):
        """
        This method applies a series of rules based on years of annotation experience
        to gene product names, attempting to correct much of the mess of things which
        which get submitted to Genbank.  This includes:

        - Removing trailing periods
        - Nonredundifies: 'protein protein', 'family family', 'family protein family protein', 
          'protein family protein' and similar
        - Changes 'superfamily' to 'family'
        - Any starting with 'ORF' or 'orf' get changed to 'conserved hypothetical protein'
        - Any products which contain 'homolog' get changed to CHP*
        - Any products with 'similar to' get changed to CHP
        - Any with 'DUF' or 'UPF' get changed to CHP
        - Any with 'ncharacteri' (captures US/british spellings of 'uncharacterized'), gets changed to CHP
        - Select words are changed to their american spellings
        - Changes any of (predicted|possible|potential|probable) to 'putative'
        - Change 'TTG start' to CHP
        - Change any starting with 'residues' to CHP
        - Removes 'C-terminus' and 'N-terminus' from end of product name
        - Strips embedded Dbxrefs from end of name:
          - Example: (2E,6E)-farnesyl diphosphate synthase {ECO:0000313|EMBL:OOP19401.1}
        - Then a long list of manual, specific name changes

        * CHP = conserved hypothetical protein

        It returns the new product name rather than overwriting the attribute.  For that,
        use set_processed_product_name().

        Note:  NCBI submission rules:  https://www.ncbi.nlm.nih.gov/genbank/asndisc.examples/
               Especially the section: SUSPECT_PRODUCT_NAMES
               These are NOT currently all implemented here
        """
        new_product = self.product_name
        default_product = 'conserved hypothetical protein'

        # remove/replace troublesome strings
        new_product = new_product.rstrip('.')
        new_product = new_product.replace('protein protein', 'protein')
        new_product = new_product.replace('superfamily', 'family')
        new_product = new_product.replace('family family', 'family')
        new_product = new_product.replace('family protein family protein', 'family protein')
        new_product = new_product.replace('family protein domain protein', 'family protein')
        new_product = new_product.replace('domain domain protein', 'domain protein')
        new_product = new_product.replace('protein family protein', 'family protein')
        new_product = new_product.replace('superfamily protein family protein', 'family protein')
        new_product = new_product.replace(' Protein-like family protein', '-like family protein')
        new_product = new_product.replace(' protein-like family protein', '-like family protein')

        # takes patterns like this off the end:  {ECO:0000313|EMBL:OOP19401.1}
        m = re.match('(.+) \{.+\:.+\}', new_product)
        if m:
            new_product = m.group(1)
        
        if new_product.lower().startswith('orf'):
            return default_product

        if 'ncharacteri' in new_product.lower():
            return default_product
            
        # process some homolog-specific names
        if 'homolog' in new_product.lower():
            if 'shiA homolog' in new_product:
                new_product = 'shiA protein'
            elif 'virulence factor mviM homolog' in new_product:
                new_product = 'virulence factor mviM'
            elif 'protein phnA homolog' in new_product:
                new_product = 'phnA protein'
            elif 'protein seqA homolog' in new_product:
                new_product = 'seqA protein'
            else:
                return default_product

        if 'similar to' in new_product.lower():
            return default_product

        if 'DUF' in new_product or 'UPF' in new_product:
            return default_product

        # If it is any form of conserved hypothetical, return the proper version of that.
        if 'onserved hypothe' in new_product.lower():
            return default_product

        if 'unnamed' in new_product.lower():
            return default_product

        # Is the name *only* protein (with whitespace)
        if new_product.lower().lstrip().rstrip() == 'protein':
            return default_product

        # Some sources give products which start with the word 'residues'
        if new_product.lower().startswith('residues'):
            return default_product

        # Some proteins are simply named 'TTG start'
        if new_product.lower().startswith('ttg start'):
            return default_product

        # Correct a class of short bogus names we've often encountered
        m = re.match('^\w{1,2}\d{1,3}$', new_product)
        if m:
            return default_product

        m = re.match('gene \d+ protein', new_product)
        if m:
            return default_product

        m = re.match('(.*)\d*\s*[CN]\-terminus$', new_product)
        if m:
            new_product = m.groups(1)

        # removes trailing symbols
        new_product = new_product.rstrip('.,-_:/')

        # Names can't end in family.  Example replacements:
        #  Raf kinase inhibitor-like protein, YbhB/YbcL family -> YbhB/YbcL family Raf kinase inhibitor-like protein
        #  phage major capsid protein, HK97 family  ->  HK97 family phage major capsid protein
        #  phage portal protein, lambda family  ->  lambda family phage portal protein
        if new_product.endswith(' family'):
            m = re.match('(.+), (.+ family)', new_product)
            if m:
                new_product = "{0} {1}".format(m.group(2), m.group(1))

        # If family still remains in the name twice, take out the first one
        #  Peptidase family S49 family protein  ->  Peptidase S49 family protein
        if new_product.count('family') > 1:
            m = re.match('(.+?) family (.+)', new_product)
            if m:
                new_product = "{0} {1}".format(m.group(1), m.group(2))

        # Americanize some words.  I'm up for arguments against these, but adding them
        #  because our previous software had them.
        new_product.replace('utilisation', 'utilization')
        new_product.replace('utilising', 'utilizing')
        new_product.replace('dimerisation', 'dimerization')
        new_product.replace('disulphide', 'disulfide')
        new_product.replace('sulphur', 'sulfur')
        new_product.replace('mobilisation', 'mobilization')

        # standardize several different forms of 'putative' to a single one
        new_product = re.sub("predicted", "putative", new_product, flags=re.I)
        new_product = re.sub("possible", "putative", new_product, flags=re.I)
        new_product = re.sub("potential", "putative", new_product, flags=re.I)
        new_product = re.sub("probable", "putative", new_product, flags=re.I)

        # Fix incorrect spellings which are getting transitively annotated
        new_product = re.sub("putaive", "putative", new_product, flags=re.I)

        # Replacements requiring word boundaries
        patt = re.compile(r'\b(?:%s)\b' % 'asparate')
        new_product = re.sub(patt, "aspartate", new_product)

        # Now a long series of manual name changes we've gathered over the years
        #  the key is the source, value is what it will be changed to.
        replacement_products = {
            'alr5027 protein': 'heme-binding protein HutZ',
            'arginine-tRNA-transferase, C terminus family protein': 'putative arginine-tRNA-transferase',
            'bacterial regulatory helix-turn-helix proteins, AraC family protein': 'transcriptional regulator, AraC family',
            'bacterial regulatory proteins, gntR family protein': 'transcriptional regulator, GntR family',
            'bacterial regulatory proteins, lacI family protein': 'transcriptional regulator, LacI family',
            'bacterial regulatory proteins, luxR family protein': 'transcriptional regulator, LuxR family',
            'bacterial regulatory proteins, tetR family protein': 'transcriptional regulator, TetR family',
            'bordetella uptake gene (bug) product family protein': 'bug family protein',
            'conserved protein with nucleoside triphosphate hydrolase domain': 'putative ATP-dependent endonuclease',
            'cyclic di-GMP binding protein VCA0042': 'cyclic di-GMP binding protein',
            'cytochrome b(C-terminal)/b6/petD family protein': 'cytochrome b family protein',
            'domain related to MnhB subunit of Na+/H+ antiporter family protein': 'Na+/H+ antiporter family protein',
            'FAD linked oxidases, C-terminal domain protein': 'FAD linked oxidase domain protein',
            'FGGY family of carbohydrate kinases, C-terminal domain protein': 'carbohydrate kinase, FGGY family',
            'gene 25-like lysozyme family protein': 'lysozyme family protein',
            'glutamate synthases, NADH/NADPH, small subunit domain protein': 'glutamate synthase, NADH/NADPH, small subunit',
            'glycogen/starch synthases, ADP-glucose type family protein': 'glycogen/starch synthase',
            'GSPII_E N-terminal domain protein': 'bacteriophage N4 adsorption protein B',
            'hydro-lases, Fe-S type, tartrate/fumarate subfamily, beta region domain protein': 'fumarate hydratase family protein',
            'invasion gene expression up-regulator, SirB family protein': 'invasion gene expression up-regulator',
            'K+ potassium transporter family protein': 'potassium uptake protein',
            'menC_gamma/gm+: o-succinylbenzoic acid (OSB) synthetase': 'o-succinylbenzoic acid (OSB) synthetase',
            'phage/plasmid replication , gene II/X family protein': 'phage/plasmid replication protein, gene II/X family',
            'phospholipase d active site motif family protein': 'phospholipase D family protein',
            'PIII': default_product,
            'putative 2-hydroxyacid dehydrogenase HI_1556': 'putative 2-hydroxyacid dehydrogenase',
            'SULFATE TRANSPORTER SULFATE TRANSPORTER FAMILY PROTEIN': 'sulfate permease family protein',
            'thiamin/thiamin pyrophosphate ABC transporter, thiamin/thiamin pyrophospate-binding protein': 'thiamin/thiamine pyrophosphate ABC transporter, thiamin/thiamine pyrophospate-binding protein',
            'traG-like , N-terminal region family protein': 'putative traG protein',
            'transcriptional activator of defense systems': 'multiple antibiotic resistance protein MarA',
            'transcriptional regulatory protein, C terminal family protein': 'putative transcriptional regulator',
            'transposase and inactivated derivative': 'putative transposase',
            'tripartite ATP-independent periplasmic transporters, DctQ component family protein': 'tripartite ATP-independent periplasmic transporter, DctQ family',
            'type IV secretory pathway VirD2 components': 'type IV secretory pathway protein',
            'zn-dependent hydrolase of the beta-lactamase fold': default_product,
        }

        for old in replacement_products:
            if new_product == old:
                new_product = replacement_products[old]
                break

        return new_product.rstrip().lstrip()
        

    def set_processed_gene_symbol(self):
        """
        See FunctionalAnnotation.process_gene_symbol() for full list of actions
        """
        self.gene_symbol = self.process_gene_symbol()

    def set_processed_product_name(self):
        """
        See FunctionalAnnotation.process_product_name() for full list of actions
        """
        self.product_name = self.process_product_name()

class Dbxref:
    """
    These allow for specification of an identifier in another database by ID.  These
    are not meant to be used for Ontology term linking, but rather limited only to
    identifiers.  Examples:
    
       SGD:S0006169
       KEGG:K06223

    The first part is the 'db' and the second is the 'identifier'.

    ## notes on method signature overloading (for future use)
    jorvis: You'd write a @classmethod. I'd name it "fromstring"
    jorvis: the classmethod would parse that string into values which it would use to
            invoke the normal constructor, as appropriate.
    """
    def __init__( self, db=None, identifier=None ):
        self.db = db
        self.identifier = identifier


class GOAnnotation:
    """
    A functional annotation can have an infinite number of associated GO Annotations

    Details here:
    http://www.geneontology.org/GO.evidence.shtml

    Yes, the 'with_from' attribute name is awkward, but 'with/from' isn't legal and
    both 'with' and 'from' are python reserved words.

    The go_id attribute here is just the numeric portion without "GO" or "GO:" or
    anything else attached (allowing the developer to define it as required.)
    """
    def __init__( self, go_id=None, ev_code=None, with_from=None ):
        self.go_id        = go_id
        self.ev_code      = ev_code
        self.with_from    = with_from

        ## process any GO ID passed to only contain the numeric portion
        go_pattern = re.compile('(\d+)')
        m = go_pattern.search(self.go_id)

        if m:
            self.go_id = m.group(1)
        else:
            raise Exception("ERROR: failed to extract numeric portion of ID from new GOAnnotation")


class ECAnnotation:
    """
    A functional annotation can have an infinite number of associated EC Annotations

    Details here:
    http://www.chem.qmul.ac.uk/iubmb/enzyme/

    While the official terms for the levels are 'class', 'subclass' etc. we have to use
    different keys for the attributes since those conflict with reserved keywords in
    both python and other frameworks.

    class1 = 1          = Oxidoreductases
    class2 = 1.10       = Acting on diphenols and related substances as donors
    class3 = 1.10.3     = With oxygen as acceptor
    number = 1.10.3.2   = laccase

    Currently does not have an index of EC terms to provide other attributes which will
    be added in the future, such as:

    accepted_name = laccase
    reaction = 4 benzenediol + O2 = 4 benzosemiquinone + 2 H2O
    systematic_name = benzenediol:oxygen oxidoreductase
    CAS_registry_number = 80498-15-3
    """
    def __init__(self, number=None):
        self.number = number
        self.class1 = None
        self.class2 = None
        self.class3 = None

        re_pattern = re.compile('(((([0-9\-]+)\.[0-9\-]+)\.[0-9\-]+)\.[a-z0-9\-]+)')
        m = re_pattern.search(self.number)

        if m:
            self.class1 = m.group(4)
            self.class2 = m.group(3)
            self.class3 = m.group(2)
            self.number = m.group(1)
        else:
            raise Exception("ERROR: Attempt to add an EC number ({0}) in unrecognized format.  Expected N.N.N.N (where N can be 0-9 or a dash)".format(self.number))
        
    def __repr__(self):
        return "EC:{0}".format(self.number)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
""" Convert a compressed :term:`FASTQ` from :term:`DSRC` to :term:`FASTQ` format"""
from bioconvert import ConvBase
import colorlog

from bioconvert.core.decorators import requires

logger = colorlog.getLogger(__name__)


__all__ = ["DSRC2GZ"]


class DSRC2GZ(ConvBase):
    """Convert a compressed :term:`FASTQ` from :term:`DSRC` to :term:`GZ` format

    Methods available are based on dsrc [DSRC]_ and pigz [PIGZ]_.

    """

    #: Default value
    _default_method = "dsrcpigz"
    _threading = True

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input DSRC filename
        :param str outfile: output GZ filename

        """
        super(DSRC2GZ, self).__init__(infile, outfile, *args, **kargs)

    @requires("dsrc")
    def _method_dsrcpigz(self, *args, **kwargs):
        """Do the conversion dsrc -> :term:`GZ`.
        Method that uses pigz and dsrc.

        `pigz documentation <https://linux.die.net/man/1/pigz>`_
        `dsrc documentation <https://github.com/refresh-bio/DSRC>`_

        option threadig does not work with the dsrc version from conda so we 
        do not add the -t threads option
        """
        cmd = "dsrc d -s  {input} | pigz -c -p {threads} > {output}"
        self.execute(
            cmd.format(threads=self.threads, input=self.infile, output=self.outfile)
        )



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################

"""Convert :term:`NEXUS` to :term:`FASTA` format"""

import colorlog

from bioconvert import ConvBase
from bioconvert.core.decorators import requires
from bioconvert.core.decorators import compressor

_log = colorlog.getLogger(__name__)


__all__ = ["NEXUS2FASTA"]


class NEXUS2FASTA(ConvBase):
    """
    Converts a sequence alignment from :term:`NEXUS` format to :term:`FASTA` format.
    """

    #: Default value
    _default_method = "biopython"

    def __init__(self, infile, outfile=None, alphabet=None, *args, **kwargs):
        """.. rubric:: constructor

        :param str infile: input :term:`NEXUS` file.
        :param str outfile: (optional) output :term:`FASTA` file
        """
        super().__init__(infile, outfile)
        self.alphabet = alphabet

    @requires("go")
    @compressor
    def _method_goalign(self, *args, **kwargs):
        """Convert :term:`NEXUS` interleaved file in :term:`FASTA` format using goalign tool.

        `goalign documentation <https://github.com/fredericlemoine/goalign>`_

        .. warning::
            the sequential format is not supported

        """
        self.install_tool("goalign")
        cmd = "goalign reformat fasta x -i {infile} -o {outfile} -x".format(
            infile=self.infile, outfile=self.outfile
        )
        self.execute(cmd)

    @requires(python_library="biopython")
    @compressor
    def _method_biopython(self, *args, **kwargs):
        """Convert :term:`NEXUS` interleaved or sequential file in :term:`FASTA` format using biopython.
        The FASTA output file will be an aligned FASTA file.

        `Bio.AlignIO <https://biopython.org/docs/1.76/api/Bio.AlignIO.html>`_

For instance:

We have a Nexus input file that look like ::

    #NEXUS
    [TITLE: Test file]

    begin data;
    dimensions ntax=3 nchar=123;
    format interleave datatype=DNA missing=N gap=-;

    matrix
    read3                -AT--------CCCGCTCGATGGGCCTCATTGCGTCCACTAGTTGATCTT
    read2                -----------------------GGAAGCCCACGCCACGGTCTTGATACG
    read4                ---------------------AGGGATGAACGATGCTCGCAGTTGATGCT

    read3                CTGGAGTAT---T----TAGGAAAGCAAGTAAACTCCTTGTACAAATAAA
    read2                AATTTTTCTAATGGCTATCCCTACATAACCTAACCGGGCATGTAATGTGT
    read4                CAGAAGTGCCATTGCGGTAGAAACAAATGTTCCCAGATTGTTGACTGATA

    read3                GATCTTA-----GATGGGCAT--
    read2                CACCGTTGTTTCGACGTAAAGAG
    read4                AGTAGGACCTCAGTCGTGACT--
    ;

    end;
    begin assumptions;
    options deftype=unord;
    end;

the output file will look like ::

    >read3
    -AT--------CCCGCTCGATGGGCCTCATTGCGTCCACTAGTTGATCTTCTGGAGTAT-
    --T----TAGGAAAGCAAGTAAACTCCTTGTACAAATAAAGATCTTA-----GATGGGCA
    T--
    >read2
    -----------------------GGAAGCCCACGCCACGGTCTTGATACGAATTTTTCTA
    ATGGCTATCCCTACATAACCTAACCGGGCATGTAATGTGTCACCGTTGTTTCGACGTAAA
    GAG
    >read4
    ---------------------AGGGATGAACGATGCTCGCAGTTGATGCTCAGAAGTGCC
    ATTGCGGTAGAAACAAATGTTCCCAGATTGTTGACTGATAAGTAGGACCTCAGTCGTGAC
    T--

and not ::

    >read3
    ATCCCGCTCGATGGGCCTCATTGCGTCCACTAGTTGATCTTCTGGAGTATTTAGGAAAGC
    AAGTAAACTCCTTGTACAAATAAAGATCTTAGATGGGCAT
    >read2
    GGAAGCCCACGCCACGGTCTTGATACGAATTTTTCTAATGGCTATCCCTACATAACCTAA
    CCGGGCATGTAATGTGTCACCGTTGTTTCGACGTAAAGAG
    >read4
    AGGGATGAACGATGCTCGCAGTTGATGCTCAGAAGTGCCATTGCGGTAGAAACAAATGTT
    CCCAGATTGTTGACTGATAAGTAGGACCTCAGTCGTGACT
"""
        from Bio import AlignIO

        with open(self.outfile, "w") as output_handle:
            alignments = list(AlignIO.parse(self.infile, "nexus"))
            AlignIO.write(alignments, output_handle, "fasta")

    @requires("squizz")
    @compressor
    def _method_squizz(self, *args, **kwargs):
        """Convert :term:`NEXUS` sequential or interleave file in :term:`FASTA` format using squizz tool.

        command used::

            squizz -c FASTA infile > outfile
        """
        cmd = "squizz -c FASTA {infile} > {outfile}".format(
            infile=self.infile, outfile=self.outfile
        )
        self.execute(cmd)




import itertools
import sys
import uuid

from bioconvert.utils.biocode import utils, gff, tbl

'''
Warning: this module requires Python 3.2 or higher

Yes, there's a lot of abstraction that could happen here still but I'm trying to allow for
special consideration for each feature type and for calls to read more like biology than CS.

Each of the 'things' is intended to be biologically relevant, and I've included the definitions
for each from the Sequence Ontology.  Their class representations may not match this perfectly
yet, but that's the goal wherever possible.  Some terms, such as Organism, fall outside the scope
of the Sequence Ontology but I found them relevant (or convenient) to include here.

Documentation should be shifted from pydoc to:
http://sphinx-doc.org/

A set of abstract classes are defined first, then the other classes, all within alphabetic order
within each group.
'''

class LocatableThing:
    '''
    Base class for any Things that are locatable on other Things, such as genes on contigs.

    You can use direct comparisons between LocatableThings to compare coordinates, such as:

        if gene1 < gene2:
            # do stuff

    These might not be what you first expect.  They are comparisons of the entire ranges
    of each feature and have the following meanings.  Keep in mind that these are all
    based on fmin and fmax coordinates, so strandedness isn't taken into account:

        thing1 > thing2 : The entire thing1 is to the right of thing2
        thing1 < thing2 : The entire thing1 is to the left of thing2
        thing1 >= thing2: The thing1 overlaps thing2 on the right end
        thing1 <= thing2: The thing1 overlaps thing2 on the left end
        thing1 == thing2: The two things share the exact same coordinates

    Here 'right' means past the fmax of the first feature and left means past the fmin.
    For visual examples, see the documentation for the functions that implement each.

    These comparisons depend on both objects sharing locations on the same object.

    '''
    def __init__( self, locations=None ):
        self.locations = locations

        if self.locations is None:
            self.locations = list()

    def __lt__(self, other):
        return self.is_on_min_side_of(thing=other)

    def __le__(self, other):
        return self.overlaps_min_side_of( thing=other )

    def __eq__(self, other):
        return self.has_same_coordinates_as(thing=other)

    def __ne__(self, other):
        return not self.has_same_coordinates_as(thing=other)

    def __gt__(self, other):
        return self.is_on_max_side_of(thing=other)

    def __ge__(self, other):
        return self.overlaps_max_side_of( thing=other )

    def contained_within( self, thing ):
        '''
        Returns True/False depending on whether the current LocatableThing is contained completely
        within the passed one.  An aligned transcript, for example, is hoped to be fully
        contained within an annotated gene.  Only the min/max bounds of the features are checked here
        currently, not internal substructures (like intron/exon bounds)

        Checks both to make sure they're both located on the same molecule before comparing coordinates,
        ignoring strandedness.
        '''
        ## see if either are located on the same thing
        for refloc in self.locations:
            for qryloc in thing.locations:
                if refloc.on.id == qryloc.on.id:
                    ## if so, see if self is completely within the passed object
                    if refloc.fmin >= qryloc.fmin and refloc.fmax <= qryloc.fmax:
                        return True

        return False

    def has_same_coordinates_as( self, on=None, thing=None, stop_tolerant=False ):
        '''
        Reports True/False depending on whether the calling object shares both start/stop coordinates
        with the passed 'thing' when considering locations 'on' the passed molecule biothing object.

        If you set the stop_tolerant argument=True, this will check the exact coordinates AND
        the coordinates with three bases off the end.  This obviously takes longer.
        '''
        other = thing

        for this_loc in self.locations:
            for other_loc in other.locations:
                # are the molecules the same?
                if this_loc.on.id == other_loc.on.id:
                    # if a specific molecule was requested, is this it?
                    if on is not None and on.id != this_loc.on.id:
                        continue

                    if this_loc.fmin == other_loc.fmin and this_loc.fmax == other_loc.fmax:
                        return True

                    # for those features with mixed stop codon inclusion
                    elif stop_tolerant == True:
                        if this_loc.strand == 1:
                            # this allows for either to be three base pairs away from the other in either direction
                            if this_loc.fmin == other_loc.fmin and abs(this_loc.fmax - other_loc.fmax) == 3:
                                return True
                        else:
                            if abs(this_loc.fmin - other_loc.fmin) == 3 and this_loc.fmax == other_loc.fmax:
                                return True

        # if we got here, there wasn't a match
        return False

    def is_on_max_side_of( self, thing=None, on=None ):
        '''
        Returns True/False depending on whether the calling LocatableThing object has a location
        relative to the passed thing that is completely past its fmax coordinate, like this:

            this :                                           fmin---------------fmax
            thing:            fmin-------------------fmax

        Checks both to make sure they're both located on the same molecule before comparing coordinates,
        ignoring strandedness.  If you pass the 'on' argument, locations will be checked against
        that specific object.
        '''
        for this_loc in self.locations:
            for other_loc in thing.locations:
                # are the molecules the same?
                if this_loc.on.id == other_loc.on.id:
                    # if a specific molecule was requested, is this it?
                    if on is not None and on.id != this_loc.on.id:
                        continue

                    if this_loc.fmin > other_loc.fmax:
                        return True

        # if we got here, there wasn't a match
        return False

    def is_on_min_side_of( self, thing=None, on=None ):
        '''
        Returns True/False depending on whether the calling LocatableThing object has a location
        relative to the passed thing that is completely before its fmin coordinate, like this:

            this :     fmin---------------fmax
            thing:                                fmin-------------------fmax

        Checks both to make sure they're both located on the same molecule before comparing coordinates,
        ignoring strandedness.  If you pass the 'on' argument, locations will be checked against
        that specific object.
        '''
        for this_loc in self.locations:
            for other_loc in thing.locations:
                # are the molecules the same?
                if this_loc.on.id == other_loc.on.id:
                    # if a specific molecule was requested, is this it?
                    if on is not None and on.id != this_loc.on.id:
                        continue

                    if this_loc.fmax < other_loc.fmin:
                        return True

        # if we got here, there wasn't a match
        return False

    def locate_on( self, target=None, fmin=None, fmin_partial=False, fmax=None, fmax_partial=False, phase=None, strand=None ):
        '''
        Locates any given biothing onto another by position range.  Use this to locate a gene
        onto an assembly, for example.  Any biothing can have multiple locations.

        Coordinates here are 0-based interbase.  Strand values are [1,0,-1].  If you pass
        "+" or "-" they'll be converted to 1 and -1, respectively.
        '''
        if strand == '+':
            strand = 1
        elif strand == '-':
            strand = -1

        loc = Location(on=target, fmin=fmin, fmin_partial=fmin_partial, fmax=fmax, fmax_partial=fmax_partial, strand=strand, phase=phase)
        self.locations.append( loc )

    def located_on( self ):
        '''
        Returns a dict of the molecules this feature is located on.  The keys are the identifiers
        and values are the objects that represent them.  This is distinct from the 'locations'
        attribute, which provides each individual location across all molecules.
        '''
        mols = dict()
        for loc in self.locations:
            mols[loc.on.id] = loc.on

        return mols

    def location( self ):
        '''
        In a majority of use-cases any given feature will only be located on one other feature.
        This method lets you access this location using a direct method call, but will raise
        an exception if called on a biothing with either 0 or more than 1 location.
        '''
        loc_count = len( self.locations )

        if loc_count == 1:
            return self.locations[0]
        elif loc_count == 0:
            raise Exception("ERROR: {0} has no locations defined".format(self.id))
        else:
            raise Exception("ERROR: {0} has more than one location defined.  You should use the location_on() method instead.".format(self.id))

    def location_on( self, mol ):
        '''
        Returns a Location object representing the location of the calling biothing on the passed
        molecule.  The name of this method implies that the object will be located on the molecule
        only a single time, and this method throws an exception if there are more than one.  If it
        isn't located on that molecule at all, None will be returned.
        '''
        loc_found = None
        for loc in self.locations:
            if loc.on.id == mol.id:
                if loc_found is None:
                    loc_found = loc
                else:
                    raise Exception("ERROR: multiple locations on the same molecule found for {0} on {1} when not expected".format(self.id, mol.id) )

        return loc_found

    def overlap_size_with( self, thing ):
        '''
        Returns None if no overlap found or not found on same molecule, else returns the number
        of overlapping bases between the two features.

        Checks both to make sure they're both located on the same molecule before comparing coordinates,
        ignoring strandedness.

        Warning: Does not check child features, which may be significant.  If you want to compare
        the exons of a transcript with the segmented match_parts of an alignment, for example
        this would only compare their overall ranges (assuming you passed transcript and match features
        for the comparison.)
        '''
        locs = self.shared_molecule_locations_with(thing)

        if locs[0] is None:
            return None

        ref_location = locs[0]
        #print("DEBUG: ref location is {0}-{1}".format(ref_location.fmin, ref_location.fmax) )
        qry_location = locs[1]
        #print("DEBUG: qry location is {0}-{1}".format(qry_location.fmin, qry_location.fmax) )

        ## first catch any which don't overlap. this makes the other tests easier
        if ref_location.fmax <= qry_location.fmin or qry_location.fmax <= ref_location.fmin:
            return None

        # overlap on right end of ref, left of qry
        # ref:   -------------------
        # qry:            -------------------
        elif ref_location.fmax <= qry_location.fmax and ref_location.fmin < qry_location.fmin:
            return ref_location.fmax - qry_location.fmin

        # overlap on left end of ref, right of qry
        # ref:           -------------------
        # qry:   -------------------
        elif ref_location.fmin >= qry_location.fmin and ref_location.fmax > qry_location.fmax:
            return qry_location.fmax - ref_location.fmin

        # is the reference contained within the qry?
        # ref:      ------------
        # qry:   -------------------
        elif ref_location.fmin >= qry_location.fmin and ref_location.fmax <= qry_location.fmax:
            return ref_location.fmax - ref_location.fmin

        # is the query contained within the ref?
        # ref:   -------------------
        # qry:      -------------
        elif qry_location.fmin >= ref_location.fmin and qry_location.fmax <= ref_location.fmax:
            return qry_location.fmax - qry_location.fmin
        else:
            raise Exception("ERROR: uncaught orientation in overlap_size_with between {0} and {1}".format(self.id, thing.id))


    def overlaps_max_side_of( self, thing=None, on=None ):
        '''
        Returns True/False depending on whether the two LocatableThing objects have
        even a single overlapping Location on the same molecule, specifically with
        respect to the fmax coordinate of the passed object.  Like this:

            this :                         fmin---------------fmax
            thing:            fmin-------------------fmax

        Checks both to make sure they're both located on the same molecule before comparing coordinates,
        ignoring strandedness.  If you pass the 'on' argument, locations will be checked against
        that specific object.
        '''
        for this_loc in self.locations:
            for other_loc in thing.locations:
                # are the molecules the same?
                if this_loc.on.id == other_loc.on.id:
                    # if a specific molecule was requested, is this it?
                    if on is not None and on.id != this_loc.on.id:
                        continue

                    if this_loc.fmin < other_loc.fmax and this_loc.fmax > other_loc.fmax:
                        return True

        # if we got here, there wasn't a match
        return False

    def overlaps_min_side_of( self, thing=None, on=None ):
        '''
        Returns True/False depending on whether the two LocatableThing objects have
        even a single overlapping Location on the same molecule, specifically with
        respect to the fmin coordinate of the passed object.  Like this:

            this :    fmin---------------fmax
            thing:            fmin-------------------fmax

        Checks both to make sure they're both located on the same molecule before comparing coordinates,
        ignoring strandedness.  If you pass the 'on' argument, locations will be checked against
        that specific object.
        '''
        for this_loc in self.locations:
            for other_loc in thing.locations:
                # are the molecules the same?
                if this_loc.on.id == other_loc.on.id:
                    # if a specific molecule was requested, is this it?
                    if on is not None and on.id != this_loc.on.id:
                        continue

                    if this_loc.fmin < other_loc.fmin and this_loc.fmax > other_loc.fmin:
                        return True

        # if we got here, there wasn't a match
        return False


    def overlaps_with( self, thing ):
        '''
        Returns True/False depending on whether the two LocatableThing objects have
        even a single overlapping Location on the same molecule.

        Checks both to make sure they're both located on the same molecule before comparing coordinates,
        ignoring strandedness.
        '''
        common_location_found = False

        ## see if either are located on the same thing
        for ref_location in self.locations:
            for qry_location in thing.locations:
                if ref_location.on.id == qry_location.on.id:
                    common_location_found = True

                    ## if so, see if they have overlapping coordinates
                    ## it's easier to negate an overlap
                    if ref_location.fmax <= qry_location.fmin or qry_location.fmax <= ref_location.fmin:
                        return False

        if common_location_found == True:
            return True
        else:
            return False

    def shared_molecule_locations_with( self, thing ):
        ## see if either are located on the same thing
        for ref_location in self.locations:
            for qry_location in thing.locations:
                if ref_location.on.id == qry_location.on.id:
                    return [ref_location, qry_location]

        ## if we got this far, one wasn't found
        return [None, None]

    def update_location( self, on=None, fmin=None, fmax=None, strand=None, phase=None ):
        '''
        Allows you to update the location of any Thing that is already located on another one.
        You can independently pass fmin, fmax, strand, or phase (or any combination) to only 
        update those attributes.
        '''
        if on is None:
            raise Exception("The 'on' attribute must be passed to the update_location() method.")

        loc = self.location_on(on)
        
        if fmin is not None: loc.fmin = fmin
        if fmax is not None: loc.fmax = fmax
        if strand is not None: loc.strand = strand
        if phase is not None: loc.phase = phase

class Location:
    '''
    Notes on the 'phase' attribute, used as defined in the GFF3 spec:

    For features of type "CDS", the phase indicates where the feature begins with reference to the
    reading frame. The phase is one of the integers 0, 1, or 2, indicating the number of bases that
    should be removed from the beginning of this feature to reach the first base of the next codon.
    In other words, a phase of "0" indicates that the next codon begins at the first base of the region
    described by the current line, a phase of "1" indicates that the next codon begins at the second
    base of this region, and a phase of "2" indicates that the codon begins at the third base of this
    region. This is NOT to be confused with the frame, which is simply start modulo 3.

    For forward strand features, phase is counted from the start field. For reverse strand features,
    phase is counted from the end field.

    The possible values for strand are 1, 0 and -1.  1 is for forward features, -1 for reverse, and
    0 is for unstranded features.

    The phase is REQUIRED for all CDS features.
    '''
    def __init__( self, on=None, fmin=None, fmin_partial=False, fmax=None, fmax_partial=False, phase=None, strand=None ):
        self.on = on
        self.fmin = fmin
        self.fmin_partial = fmin_partial
        self.fmax = fmax
        self.fmax_partial = fmax_partial
        self.strand = strand
        self.phase = 0 if phase is None else phase
        

class MoleculeSet:
    '''
    Base class for any Set of different molecule types, such as AssemblySet or PolypeptideSet.

    It is appropriate for use for all classes inheriting from LocatableThing.
    '''
    def __init__(self):
        # Nothing here yet
        pass

    def load_from_dict(self, thedict):
        for id in thedict:
            self.add(thedict[id])

    def write_fasta(self, fh=None, path=None):
        '''
        Writes the current set in FASTA format.  You can either pass the fh or path arguments.  If
        an open file handle already exists, fh is appropriate.  Instead, if you have just a path you
        want to be written to pass the 'path' argument instead.

        The header format in the FASTA entries depends on the type of elements in the set.  
        '''
        if path is not None:
            fh = open(path, 'wt')

        if self.__class__ == PolypeptideSet:
            molecules = self.polypeptides
        elif self.__class__ == AssemblySet:
            molecules = self.assemblies
        else:
            raise Exception("ERROR: writing FASTA not supported in MoleculeSets of this type: {0}".format(self.__class__))

        for molecule in molecules:
            if self.__class__ == PolypeptideSet:
                header = molecule.annotation_string()
            elif self.__class__ == AssemblySet:
                header = molecule.id
                
            fh.write(">{0}\n".format(header))
            fh.write("{0}\n".format(utils.wrapped_fasta(molecule.residues)))

        fh.close()
        
class Organism:
    '''
    This is a pretty high-level representation of an organism, and (currently) mostly serves as a
    container for Assembly objects.
    '''
    def __init__( self, id=None, assemblies=None, taxon_id=None, genus=None, species=None, strain=None ):
        self.id = id
        self.assemblies = assemblies
        self.taxon_id = taxon_id

        # convenience attributes
        self.genus = genus
        self.species = species
        self.strain = strain

        if self.assemblies is None:
            self.assemblies = list()

        def add_assembly( self, assembly ):
            self.assemblies.append( assembly )

        def common_name(self):
            if self.strain is None:
                return "{0} {1}".format(self.genus, self.species)
            else:
                return "{0} {1} ({2})".format(self.genus, self.species, self.strain)



class Assembly( LocatableThing ):
    '''
    SO definition (2013-05-22): "A region of the genome of known length that is composed by ordering and
    aligning two or more different regions."
    '''

    def __init__( self, id=None, locations=None, length=None, residues=None, is_circular=None, children=None ):
        super().__init__(locations)
        self.id = id
        self.length = length
        self.residues = residues
        self.is_circular = is_circular
        self.children = children

        if length is None and self.residues is not None and len(self.residues) > 0:
            self.length = len(self.residues)

        ## initialize any types needed
        self.children = _initialize_type_list(self.children, 'gene')

    def add_gene( self, gene ):
        self.children['gene'].append(gene)

    def genes(self):
        '''
        Returns all Gene objects which are children of this Assembly
        '''
        return self.children['gene']


class AssemblySet( MoleculeSet ):
    '''
    This is a convenience class for when operations need to be performed on a set of contigs/scaffolds
    such as N50 calculation or writing to a FASTA file.
    '''
    def __init__( self, assemblies=None ):
        super().__init__()
        self.assemblies = assemblies

        if self.assemblies is None:
            self.assemblies = list()

    def add( self, assembly ):
        self.assemblies.append(assembly)

    def load_from_file(self, file):
        seqs = utils.fasta_dict_from_file(file)
        
        for seq_id in seqs:
            assembly = Assembly(id=seq_id, residues=seqs[seq_id]['s'])
            self.add(assembly)

    def N50( self ):
        '''
        N50 is a statistical measure of average length of a set of sequences. It is used widely in
        genomics, especially in reference to contig or supercontig lengths within a draft assembly.

        Given a set of sequences of varying lengths, the N50 length is defined as the length N for
        which 50% of all bases in the sequences are in a sequence of length L < N. This can be found
        mathematically as follows: Take a list L of positive integers. Create another list L' ,
        which is identical to L, except that every element n in L has been replaced with n copies of
        itself. Then the median of L' is the N50 of L. For example: If L = {2, 2, 2, 3, 3, 4, 8, 8},
        then L' consists of six 2's, six 3's, four 4's, and sixteen 8's; the N50 of L is the median
        of L' , which is 6.
        
        Source: http://www.broadinstitute.org/crd/wiki/index.php/N50
        '''
        total_len = 0
        lengths = list()

        for assembly in self.assemblies:
            total_len += assembly.length
            lengths.append(assembly.length)

        lengths = sorted(lengths)

        count = 0

        for i in range(len(lengths)):
            count += lengths[i]
            if count >= total_len / 2:
                return lengths[i]


class CDS( LocatableThing ):
    '''
    SO definition (2013-05-22): "A contiguous sequence which begins with, and includes, a start codon
    and ends with, and includes, a stop codon."

    Note that there is a minor variation from this, in that there is no enforcement or expectation
    that the sequence coordinates or residues stored begin with a start or end with a stop.  This is
    so that we can still use this object to represent CDS within partial gene predictions, which may
    or may not have complete ends.
    '''
    def __init__( self, id=None, locations=None, parent=None, phase=None, length=None, residues=None, annotation=None ):
        super().__init__(locations)
        self.id = id
        self.parent = parent
        self.length = length
        self.residues = residues
        self.phase = 0 if phase is None else phase

        ## this should be an instance of FunctionalAnnotation from annotation.py
        self.annotation = annotation

    def get_residues(self):
        if len(self.locations) == 0:
            raise Exception("ERROR: CDS.get_residues() requested but CDS {0} isn't located on anything.".format(self.id))
        elif len(self.locations) > 1:
            raise Exception("ERROR: CDS {0} is located on multiple molecules.  Can't automatically extract the residues.".format(self.id))

        loc = self.location()
        mol = loc.on

        # make sure this thing has its residues populated
        if len(mol.residues) <= 0:
            raise Exception("ERROR: CDS.get_residues() requested but its molecule {0} has no stored residues".format(mol.id))

        self.residues = mol.residues[loc.fmin:loc.fmax]
        self.length = len(self.residues)

        if loc.strand == -1:
            self.residues = utils.reverse_complement(self.residues)

        return self.residues


class Exon( LocatableThing ):
    def __init__( self, id=None, locations=None, parent=None, length=None ):
        super().__init__(locations)
        self.id = id
        self.parent = parent
        self.length = length


class Gene( LocatableThing ):
    '''
    SO definition (2013-05-22): "A region (or regions) that includes all of the sequence
    elements necessary to encode a functional transcript. A gene may include regulatory
    regions, transcribed regions and/or other functional sequence regions."
    '''
    def __init__( self, id=None, locus_tag=None, locations=None, children=None, residues=None ):
        super().__init__(locations)
        self.id = id
        self.locus_tag = locus_tag
        self.children = children
        self.residues = residues

        ## initialize any types needed
        self.children = _initialize_type_list(self.children, 'mRNA')
        self.children = _initialize_type_list(self.children, 'rRNA')
        self.children = _initialize_type_list(self.children, 'tRNA')
        self.children = _initialize_type_list(self.children, 'tmRNA')

    def __hash__(self):
        return hash(self.id)

    def add_RNA(self, rna):
        """
        Utility method which can be used to add any RNA subtype (mRNA, rRNA, etc.)
        """
        if isinstance(rna, mRNA):
            self.add_mRNA(rna)
        elif isinstance(rna, rRNA):
            self.add_rRNA(rna)
        elif isinstance(rna, tRNA):
            self.add_tRNA(rna)
        elif isinstance(rna, tmRNA):
            self.add_tmRNA(rna)
        else:
            raise Exception("ERROR: add_RNA() method called for unrecognized type: {0}".format(rna.__class__.__name__))

    def add_mRNA(self, rna):
        rna.parent = self
        self.children['mRNA'].append(rna)

    def add_rRNA(self, rna):
        rna.parent = self
        self.children['rRNA'].append(rna)

    def add_tRNA(self, rna):
        rna.parent = self
        self.children['tRNA'].append(rna)

    def add_tmRNA(self, rna):
        rna.parent = self
        self.children['tmRNA'].append(rna)

    def get_residues(self):
        if len(self.locations) == 0:
            raise Exception("ERROR: gene.get_residues() requested but gene {0} isn't located on anything.".format(self.id))
        elif len(self.locations) > 1:
            raise Exception("ERROR: gene {0} is located on multiple molecules.  Can't automatically extract the residues.".format(self.id))

        loc = self.location()
        mol = loc.on

        # make sure this thing has its residues populated
        if len(mol.residues) <= 0:
            raise Exception("ERROR: gene.get_residues() requested but its molecule {0} has no stored residues".format(mol.id))

        self.residues = mol.residues[loc.fmin:loc.fmax]
        self.length = len(self.residues)

        if loc.strand == -1:
            self.residues = utils.reverse_complement(self.residues)

        return self.residues

    def mRNA_count(self):
        return len(self.mRNAs())

    def mRNAs(self):
        return self.children['mRNA']

    def ncRNAs(self):
        return self.children['tRNA'] + self.children['rRNA'] + self.children['tmRNA']
    
    def polypeptides(self):
        """
        Returns a list() of all Polypeptide objects for all mRNA children.
        """
        polypeptides = list()

        for rna in self.RNAs():
            polypeptides.extend(rna.polypeptides())

        return polypeptides
        

    def RNAs(self):
        return self.children['mRNA'] + self.children['tRNA'] + self.children['rRNA'] + self.children['tmRNA']

    def remove_mRNA(self, rna):
        if rna in self.children['mRNA']:
            self.children['mRNA'].remove(rna)
        
    def rRNAs(self):
        return self.children['rRNA']

    def shares_exon_structure_with( self, thing=None, stop_tolerant=False ):
        """
        This checks if two genes have only one mRNA and, if so, compares their internal
        exon structure.  Returns True if they all match completely.
        """
        these_mRNAs = self.mRNAs()
        other_mRNAs = thing.mRNAs()

        if len(these_mRNAs) != 1 or len(other_mRNAs) != 1 or len(these_mRNAs[0].exons()) != len(other_mRNAs[0].exons()):
            return False;

        exon_matches_found = 0
        ref_exon_count = 0

        for ref_exon in these_mRNAs[0].exons():
            ref_exon_count += 1

            for other_exon in other_mRNAs[0].exons():
                if ref_exon.has_same_coordinates_as( thing=other_exon, stop_tolerant=stop_tolerant ):
                    exon_matches_found += 1
                    break

        if exon_matches_found == ref_exon_count:
            return True
        else:
            return False

    def shares_CDS_structure_with( self, thing=None ):
        """
        This checks if two genes have only one mRNA and, if so, compares their internal
        CDS structure.  Returns True if they all match completely.
        """
        these_mRNAs = self.mRNAs()
        other_mRNAs = thing.mRNAs()

        if len(these_mRNAs) != 1 or len(other_mRNAs) != 1 or len(these_mRNAs[0].CDSs()) != len(other_mRNAs[0].CDSs()):
            return False;

        CDS_matches_found = 0
        ref_CDS_count = 0

        for ref_CDS in these_mRNAs[0].CDSs():
            ref_CDS_count += 1

            for other_CDS in other_mRNAs[0].CDSs():
                if ref_CDS.has_same_coordinates_as( thing=other_CDS ):
                    CDS_matches_found += 1
                    break

        if CDS_matches_found == ref_CDS_count:
            return True
        else:
            return False

    def tRNAs(self):
        return self.children['tRNA']

    def print_as(self, fh=None, source=None, format=None, lab=None):
        '''
        Prints the gene entry in a variety of available formats, with all children.
        Is printed to STDOUT unless the fh option is passed.  Really, only checks are
        done here before passing the info to the biocodegff module.

        The options passed will decide the target and format of the print.

            --fh = File handle to which we should print (default: STDOUT)
            --format = Currently only 'gff3', 'tbl' or the default 'text'.
            --source = Required for gff printing
            --lab = Required for tbl printing
        ''' 
        if format == 'text':
            _print_thing(self, fh=fh)
        elif format == 'gff3':
            gff.print_biogene(gene=self, fh=fh, source=source)
        elif format == 'tbl':
            tbl.print_biogene(gene=self, fh=fh, lab=lab)
        else:
            raise Exception("ERROR: the print_as method only accepts values of 'text' or 'gff3'")

    def remove_mRNA(self, rna):
        if rna in self.children['mRNA']:
            self.children['mRNA'].remove(rna)
        else:
            raise Exception("ERROR: Can't remove mRNA {0} from gene {1} because it wasn't found".format(rna.id, self.id))


class Intron( LocatableThing ):
    '''
    Usually not represented directly.  Rather, these are generated on the fly
    as needed when the RNA.introns() method is called.

    There are some issues to resolve here.  For example, what's the parentage
    of an intron?  What are conceivable children?

    SO definition (2013-05-22): "A region of a primary transcript that is transcribed, but
    removed from within the transcript by splicing together the sequences (exons) on either
    side of it."
    '''
    def __init__( self, id=None, locations=None, length=None ):
        super().__init__(locations)
        self.id = id
        self.length = length


class Match( LocatableThing ):
    '''
    Represents a generic match object, usually used to represent pairwise alignments.
    The sequence ontology provides a host of specific match terms, but I'm using this
    as a lightweight implementation of them all rather than explicitly creating
    dozens of subclasses.  Instead, create a Match object and simply set any more
    specific term using Match.subclass

    Future mod: Add Annotation object to store info on the match?
    '''

    def __init__( self, id=None, subclass=None, target_id=None, locations=None, parts=None, length=None ):
        super().__init__(locations)
        self.id = id
        self.subclass = subclass   # allows you to define specific classes (cDNA_match, etc.)
        self.target_id = target_id
        self.parts = parts   # these are MatchPart objects
        self.length = length

        if self.subclass is None:
            self.subclass = 'match'

        if self.parts is None:
            self.parts = list()

    #def __repr__(self):
    #    '''
    #    Prints a string representation of the match along with it's match_parts
    #    '''
    #    lines = list()
    #    lines.append("Match - id:{0}, subclass:{1}, target_id:{2}, length:{3}".format( \
    #                  self.id, self.subclass, self.target_id, self.length))
    #    lines.append("\tlocations: {0}".format(len(self.locations)) )
    #    lines.append("\tmatch_parts: {0}".format(len(self.parts)) )

    #    for mp in self.parts:
    #        print(mp)
    #        lines.append("\t\tMatchPart - id:{0}, parent_id:{1}, length{2}".format( \
    #                  mp.id, mp.parent.id, mp.length))
    #        lines.append("\t\t\tlocations: {0}".format(len(self.locations)) )

    #    return "\n".join(lines)

    def add_part(self, part):
        self.parts.append(part)

    def print_as(self, fh=None, source=None, format=None):
        if format == 'text':
            _print_thing(self, fh=fh)
        elif format == 'gff3':
            # mode could be passed here to print both GFF3-supported representations (already implemented)
            gff.print_biomatch(match=self, fh=fh, source=source)
        else:
            raise Exception("ERROR: You attempted to print a Match with unrecognized format:{0}".format(format))


class MatchPart( LocatableThing ):
    """ Future mod: Add Annotation object to store info on the match? """
    def __init__( self, id=None, locations=None, parent=None, length=None ):
        super().__init__(locations)
        self.id = id
        self.parent = parent  # this should be a Match object
        self.length = length

    #def __repr__(self):
    #    '''
    #    Prints a string representation of the match_part along with it's locations
    #    '''
    #    lines = list()
    #    lines.append("MatchPart - id:{0}, parent_id:{1}, length{2}".format( \
    #                  self.id, self.parent.id, self.length))
    #    lines.append("\tlocations: {0}".format(len(self.locations)) )

    #    return "\n".join(lines)

    
class Polypeptide( LocatableThing ):
    '''
    SO definition (2013-05-22): "A sequence of amino acids linked by peptide bonds which may lack
    appreciable tertiary structure and may not be liable to irreversible denaturation."
    '''
    def __init__( self, id=None, locations=None, parent=None, length=None, annotation=None, residues=None ):
        super().__init__(locations)
        self.id = id
        self.parent = parent
        self.length = length
        self.residues = residues

        ## this should be an instance of FunctionalAnnotation from annotation.py
        self.annotation = annotation

    def annotation_string(self):
        '''
        Returns a string of the annotation attached to this Polypeptide, most often used as
        a FASTA entry header
        '''
        header = ""
        go_string = ""
        ec_string = ""

        if self.annotation is None:
            return self.id
        else:
            for go_annot in self.annotation.go_annotations:
                go_string += "GO:{0},".format(go_annot.go_id)

            go_string = go_string.rstrip(',')

            for ec_annot in self.annotation.ec_numbers:
                ec_string += "{0},".format(ec_annot.number)

            ec_string = ec_string.rstrip(',')

            header = "{0} {1}".format(self.id, self.annotation.product_name)

            if self.annotation.gene_symbol is not None:
                header = "{0} gene::{1}".format(header, self.annotation.gene_symbol)

            if ec_string != "":
                header = "{0} ec::{1}".format(header, ec_string)

            if go_string != "":
                header = "{0} go::{1}".format(header, go_string)

        return header

    
class PolypeptideSet( MoleculeSet ):
    '''
    This is a convenience class for when operations need to be performed on a set of polypeptides
    such as writing to a FASTA file.
    '''
    def __init__( self, polypeptides=None ):
        super().__init__()
        self.polypeptides = polypeptides

        if self.polypeptides is None:
            self.polypeptides = list()

    def add( self, polypeptide ):
        self.polypeptides.append( polypeptide )

    def load_from_file(self, file):
        seqs = utils.fasta_dict_from_file(file)
        
        for seq_id in seqs:
            polypeptide = Polypeptide(id=seq_id, residues=seqs[seq_id]['s'])
            self.add(polypeptide)

class Region( LocatableThing ):
    '''
    So definition (2020-08-31): "A sequence_feature with an extent greater than zero. A nucleotide 
    region is composed of bases and a polypeptide region is composed of amino acids."
    '''
    def __init__( self, id=None, locations=None ):
        super().__init__(locations)
        self.id = id
            
class RNA( LocatableThing ):
    '''
    SO definition (2013-05-22): "An attribute describing a sequence consisting of nucleobases
    bound to a repeating unit made of a D-ribose ring connected to a phosphate backbone."
    '''
    def __init__( self, id=None, locations=None, parent=None, locus_tag=None, children=None, annotation=None ):
        super().__init__(locations)
        self.id = id
        self.parent = parent
        self.locus_tag = locus_tag
        self.children = children

        ## This should be an instance of FunctionalAnnotation from annotation.py
        #   It's considered best practice to put the annotation on the Polypeptide feature for
        #   coding RNAs.
        #   Here we expect things like product (rRNAs), anticodon (tRNAs), etc.
        self.annotation = annotation

        if annotation is not None:
            print("Warning: The RNA.annotation attribute is deprecated and will soon be removed.  Instead, use ncRNA.annotation", file=sys.stderr)

        ## initialize any types needed
        self.children = _initialize_type_list(self.children, 'exon')
        self.children = _initialize_type_list(self.children, 'CDS')
        self.children = _initialize_type_list(self.children, 'polypeptide')
        self.children = _initialize_type_list(self.children, 'UTR')

    def __hash__(self):
        return hash(self.id)

    def add_CDS(self, cds):
        cds.parent = self
        self.children['CDS'].append(cds)

    def add_exon(self, exon):
        exon.parent = self
        self.children['exon'].append(exon)

    def add_five_prime_UTR(self, utr):
        utr.parent = self
        self.children['UTR'].append(utr)

    def add_polypeptide(self, polypeptide):
        polypeptide.parent = self
        self.children['polypeptide'].append(polypeptide)

    def add_three_prime_UTR(self, utr):
        utr.parent = self
        self.children['UTR'].append(utr)

    def add_UTR(self, utr):
        if isinstance(utr, FivePrimeUTR):
            self.add_five_prime_UTR(utr)
        elif isinstance(utr, ThreePrimeUTR):
            self.add_three_prime_UTR(utr)
        else:
            raise Exception("ERROR: add_UTR() method called for unrecognized type: {0}".format(utr.__class__.__name__))

    def CDS_count(self):
        return len(self.CDSs())

    def CDSs(self):
        ## Why not "CDSes"?  As a grammarian, this gave me fits.  There are many references
        #  which suggest adding -es to any initialism, but in the end I had to go with Oxford's example:
        #  http://oxforddictionaries.com/definition/american_english/SOS
        #   -- J. Orvis
        #
        # This method was added for consistency of usage and naming, but it really returns an array
        #  of the CDS fragments in spliced genes.  If you want the contiguous sequence, use get_CDS_residues()
        #
        # Note that this is not guaranteed to return the CDSs in sorted chromosomal order
        return self.children['CDS']

    def delete_CDS(self, cds_to_remove):
        idx = 0

        for cds in self.children['CDS']:
            if cds.id == cds_to_remove.id:
                del self.children['CDS'][idx]
                break

            idx += 1
        else:
            raise Exception("ERROR: Attempt to delete a CDS ({0}) which wasn't found as a child of this RNA ({1}".format(cds.id, self.id))

    def exon_count(self):
        return len(self.exons())

    def exons(self):
        return self.children['exon']

    def extend_stop(self, on=None, to=None):
        '''
        Extending an RNA involves changing its coordinates, checking if the extension requires
        the parent gene to be extended, and checking sub features (exon, CDS) for extension.
        This does all of those things with respect to the molecule passed via the 'on' argument.

        on = Molecule object this RNA is located on
        to = To what coordinate should the stop be extended?  The strand will be checked
             automatically
        '''
        if on is None or to is None:
            raise Exception("The 'on' and 'to' attributes must be passed to the extend_stop() method.")

        if to < 0:
            raise Exception("Attempt to extend coordinates for {0} on {1} to a negative coordinate ({2})".format(self.id, on.id, to))

        # Do we need to extend the gene?
        gene = self.parent
        gene_loc = gene.location_on(on)
        self_loc = self.location_on(on)

        if self_loc.strand == 1:
            # We don't always need to extend the RNA.  It may just be the internal structure that needs it.
            if self_loc.fmax < to:
                self_loc.fmax = to

            if gene_loc.fmax < to:
                gene_loc.fmax = to

            if self.CDS_count() > 0:
                terminal_CDS = sorted(self.CDSs())[-1]
                terminal_CDS_loc = terminal_CDS.location_on(on)
                terminal_CDS_loc.fmax = to

            if self.exon_count() > 0:
                terminal_exon = sorted(self.exons())[-1]
                terminal_exon_loc = terminal_exon.location_on(on)
                terminal_exon_loc.fmax = to
        else:
            if self_loc.fmin > to:
                self_loc.fmin = to                

            if gene_loc.fmin > to:
                gene_loc.fmin = to

            if self.CDS_count() > 0:
                terminal_CDS = sorted(self.CDSs())[0]
                terminal_CDS_loc = terminal_CDS.location_on(on)
                terminal_CDS_loc.fmin = to

            if self.exon_count() > 0:
                terminal_exon = sorted(self.exons())[0]
                terminal_exon_loc = terminal_exon.location_on(on)
                terminal_exon_loc.fmin = to
        
    
    def five_prime_UTRs(self):
        utrs = list()

        for utr in self.UTRs():
            if isinstance(utr, FivePrimeUTR):
                utrs.append(utr)

        return utrs

    def get_CDS_residues(self, for_translation=False):
        '''
        Returns a DNA string representing the CDS based on the coordinate spans.  Note that phase
        will NOT be respected unless you pass for_translation=True
        '''
        if len(self.locations) == 0:
            raise Exception("ERROR: RNA.get_CDS_residues() requested but RNA {0} isn't located on anything.".format(self.id))
        elif len(self.locations) > 1:
            raise Exception("ERROR: RNA {0} is located on multiple molecules.  Can't automatically extract the residues.".format(self.id))

        loc = self.location()
        segments = list()

        chop = None

        # these MUST be handled in a sorted order
        residues = ''
        sorted_cds = sorted(self.CDSs())
        
        if len(sorted_cds) > 0:
            for cds in sorted_cds:
                cds_segment = cds.get_residues()
                segments.append(cds_segment)

            if loc.strand == -1:
                segments = reversed(segments)
                chop = sorted_cds[-1].phase
            else:
                chop = sorted_cds[0].phase

            for segment in segments:
                residues += segment

            if chop is not None:
                residues = residues[chop:]

        return residues

    def has_introns( self ):
        if len( self.exons() ) > 1:
            return True
        else:
            return False

    def introns(self, on=None):
        '''
        Dynamically generates Intron objects in order for the current RNA.  The coordinates of the
        generated introns depend on the object passed via the 'on' argument
        '''
        if on is None:
            raise Exception("ERROR: the introns() method requires a passed molecule using the 'on' argument")

        mol_on = on

        intron_objs = list()
        last_exon = None
        last_exon_loc = None

        for exon in sorted(self.exons()):
            exon_loc = exon.location_on( mol_on )

            if last_exon is not None:
                intron_id = uuid.uuid4()
                intron = Intron( id=intron_id )
                intron.locate_on( target=mol_on, fmin=last_exon_loc.fmax, fmax=exon_loc.fmin, strand=exon_loc.strand )
                intron_objs.append( intron )

            last_exon = exon
            last_exon_loc = exon_loc

        return intron_objs

    def polypeptides(self):
        return self.children['polypeptide']

    def three_prime_UTRs(self):
        utrs = list()

        for utr in self.UTRs():
            if isinstance(utr, ThreePrimeUTR):
                utrs.append(utr)

        return utrs

    def UTRs(self):
        return self.children['UTR']

class mRNA( RNA ):
    '''
    SO definition (2013-05-22): "Messenger RNA is the intermediate molecule between DNA and protein. It
    includes UTR and coding sequences. It does not contain introns."
    '''
    def __init__( self, id=None, locations=None, parent=None, locus_tag=None, children=None, residues=None ):
        super().__init__(id, locations, parent, locus_tag, children)
        self.residues = residues

    def longest_orf(self, require_start=False, frame=None):
        '''
        Searches all six frames (unless a specific frame is passed) and returns an 
        open-reading frame (ORF) instance.

        If you want to require a start codon pass require_start=True and the ORF will be 
        checked for one of AUG/ATG, GUG/GTG or UUG/TTG.  

        If the longest ORF is on the reverse strand the residues within the ORF object
        are reverse complemented.
        '''
        if self.residues is None:
            raise Exception("mRNA.longest_orf() called but mRNA.residues was empty")
        
        stops = ["TAG", "TGA", "TAA"]
        starts = ["ATG", "GTG", "TTG"]
        
        if frame is None:
            frames = [1,2,3,-1,-2,-3]
        elif frame not in [1,2,3,-1,-2,-3]:
            raise Exception("Invalid value for frames.  mRNA.longest_orf() only considers one of [1,2,3,-1,-2,-3]")
        else:
            frames = [frame,]

        longest_residues = ''
        longest_frame = None

        working_seq = self.residues.upper()
        working_seq.replace('U', 'T')
        working_seq_revcomped = False

        for frame in frames:
            real_frame = frame
            
            if frame < 0:
                if not working_seq_revcomped:
                    working_seq = utils.reverse_complement(working_seq)
                    working_seq_revcomped = True
                frame = 0 - frame

            current_residues = ''
                
            for i in range(frame - 1, len(working_seq), 3):
                codon = working_seq[i:i+3]
                ## if the current codon is a stop, record this if it's the longest and restart
                if codon in stops:
                    if len(current_residues) > len(longest_residues):
                        if require_start:
                            if current_residues[0:3] in starts:
                                longest_residues = current_residues
                                longest_frame = real_frame
                        else:
                            longest_residues = current_residues
                            longest_frame = real_frame

                    current_residues = ''
                else:
                    current_residues += codon

            # handle the last case
            if len(current_residues) > len(longest_residues):
                if require_start:
                    if current_residues[0:3] in starts:
                        longest_residues = current_residues
                        longest_frame = real_frame
                else:
                    longest_residues = current_residues
                    longest_frame = real_frame

        return ORF(parent=self, frame=longest_frame, residues=longest_residues)

class mRNASet( MoleculeSet ):
    '''
    This is a convenience class for when operations need to be performed on a set of mRNAs
    such as writing to a FASTA file.
    '''
    def __init__( self, mRNAs=None ):
        super().__init__()
        self.mRNAs = mRNAs

        if self.mRNAs is None:
            self.mRNAs = list()

    def add( self, mRNA ):
        self.mRNAs.append( mRNA )

    def load_from_file(self, file):
        seqs = utils.fasta_dict_from_file(file)
        
        for seq_id in seqs:
            mRNA = mRNA(id=seq_id, residues=seqs[seq_id]['s'])
            self.add(mRNA)

class ncRNA( RNA ):
    '''
    SO definition (2019-04-17): "An RNA transcript that does not encode for a protein rather the RNA 
    molecule is the gene product."

    An ncRNA object has an annotation associated with it, since for protein-coding genes it is added 
    on the polypeptide level.
    '''
    def __init__( self, id=None, locations=None, parent=None, locus_tag=None, children=None, annotation=None ):
        super().__init__(id, locations, parent, locus_tag, children)

        ## this should be an instance of FunctionalAnnotation from annotation.py
        self.annotation = annotation

class ORF( LocatableThing ):
    '''
    SO definition (2020-04-14): "The in-frame interval between the stop codons of a reading frame 
    which when read as sequential triplets, has the potential of encoding a sequential string of 
    amino acids. TER(NNN)nTER."

    Note that this doesn't mean the sequence begins with a start codon.  It's simply devoid of 
    stops.

    The parent attribute should point to an mRNA instance.
    '''
    def __init__(self, id=None, locations=None, parent=None, frame=None, residues=None):
        super().__init__(locations)
        self.parent = parent
        self.frame = frame
        self.residues = residues
            
class tmRNA( ncRNA ):
    '''
    SO definition (2017-11-29): "A tmRNA liberates a mRNA from a stalled ribosome. To accomplish this part 
    of the tmRNA is used as a reading frame that ends in a translation stop signal. The broken mRNA is 
    replaced in the ribosome by the tmRNA and translation of the tmRNA leads to addition of a proteolysis 
    tag to the incomplete protein enabling recognition by a protease. Recently a number of permuted tmRNAs 
    genes have been found encoded in two parts. tmRNAs have been identified in eubacteria and some 
    chloroplasts but are absent from archeal and Eukaryote nuclear genomes."
    '''
    def __init__( self, id=None, locations=None, parent=None, locus_tag=None, children=None ):
        super().__init__(id, locations, parent, locus_tag, children)
            
class rRNA( ncRNA ):
    '''
    SO definition (2013-05-22): "RNA that comprises part of a ribosome, and that can provide both
    structural scaffolding and catalytic activity."
    '''
    def __init__( self, id=None, locations=None, parent=None, locus_tag=None, children=None, annotation=None ):
        super().__init__(id, locations, parent, locus_tag, children)

        ## this should be an instance of FunctionalAnnotation from annotation.py
        self.annotation = annotation

class tRNA( ncRNA ):
    '''
    SO definition (2013-05-22): "Transfer RNA (tRNA) molecules are approximately 80 nucleotides in length.
    Their secondary structure includes four short double-helical elements and three loops (D, anti-codon,
    and T loops). Further hydrogen bonds mediate the characteristic L-shaped molecular structure. Transfer
    RNAs have two regions of fundamental functional importance: the anti-codon, which is responsible for
    specific mRNA codon recognition, and the 3' end, to which the tRNA's corresponding amino acid is
    attached (by aminoacyl-tRNA synthetases). Transfer RNAs cope with the degeneracy of the genetic code
    in two manners: having more than one tRNA (with a specific anti-codon) for a particular amino acid;
    and 'wobble' base-pairing, i.e. permitting non-standard base-pairing at the 3rd anti-codon position."
    '''
    def __init__( self, id=None, locations=None, parent=None, locus_tag=None, children=None, anticodon=None ):
        super().__init__(id, locations, parent, locus_tag, children)
        self.anticodon = anticodon

class UTR( LocatableThing ):
    '''
    SO definition (2014-06-12): "Messenger RNA sequences that are untranslated and lie five prime or three
    prime to sequences which are translated."
    '''
    def __init__( self, id=None, locations=None, parent=None, length=None ):
        super().__init__(locations)
        self.id = id
        self.parent = parent
        self.length = length

class FivePrimeUTR( UTR ):
    '''
    SO definition (2014-06-12): "A region at the 5' end of a mature transcript (preceding the initiation
    codon) that is not translated into a protein."
    '''
    def __init__( self, id=None, locations=None, parent=None, length=None ):
        super().__init__(id, locations, parent, length)

class ThreePrimeUTR( UTR ):
    '''
    SO definition (2014-06-12): "A region at the 3' end of a mature transcript (following the stop codon)
    that is not translated into a protein."
    '''
    def __init__( self, id=None, locations=None, parent=None, length=None ):
        super().__init__(id, locations, parent, length)


#############################
## Private help functions
##   Hopefully this isn't too un-pythonic.  These were to aid in abstraction within classes
#############################

def _initialize_type_list( children, feattype ):
    if children is None:
        children = dict()

    if feattype not in children:
        children[feattype] = list()

    return children


def _print_thing( thing, fh=None ):
    '''
    This is mostly used for debugging.  Pass any of the biothings like Gene, CDS, etc
    and this will print out its attributes including locations, children info, etc.
    '''
    print("DEBUG: Info for biothing with ID:({0})".format( thing.id ) )

    if hasattr(thing, 'parent'):
        print("\tparent = {0}".format(thing.parent.id) )
    else:
        print("\tparent = None")

    if hasattr(thing, 'children'):
        child_count = 0

        for child_type in thing.children:
            for child in thing.children[child_type]:
                child_count += 1
                print("\t\t{0}".format(child.id) )

        print("\tchild count = {0}".format( child_count ) )
    else:
        print("\tchildren = None")

    if hasattr(thing, 'locations'):
        print("\tlocation count: {0}".format(len(thing.locations)) )
        print("\tlocations:")

        for loc in thing.locations:
            print("\t\t{0}\t{1}\t{2}\t{3}".format(loc.on.id, loc.fmin, loc.fmax, loc.strand) )
    else:
        print("\tlocations = None")



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

_MAJOR = 0
_MINOR = 5
_MICRO = 1
version = "%d.%d.%d" % (_MAJOR, _MINOR, _MICRO)
release = "%d.%d" % (_MAJOR, _MINOR)

metainfo = {
    "authors": {"Cokelaer": ("Thomas Cokelaer", "thomas.cokelaer@pasteur.fr"),},
    "version": version,
    "license": "GPLv3",
    "download_url": "https://github.com/bioconvert/bioconvert",
    "url": "https://github.com/bioconvert/bioconvert",
    "description": "convert between bioinformatics formats",
    "platforms": ["Linux", "Unix", "MacOsX", "Windows"],
    "keywords": ["NGS", "bam2bed", "fastq2fasta", "bam2sam"],
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
}


with open("README.rst") as f:
    readme = f.read()

requirements = open("requirements.txt").read().split()
on_rtd = os.environ.get("READTHEDOCS", None) == "True"
if on_rtd:
    # mock, pillow, sphinx, sphinx_rtd_theme installed on RTD
    # but we also need numpydoc and sphinx_gallery
    extra_packages = ["numpydoc", "sphinx_gallery"]
    requirements += extra_packages


setup(
    name="bioconvert",
    version=version,
    maintainer=metainfo["authors"]["Cokelaer"][0],
    maintainer_email=metainfo["authors"]["Cokelaer"][1],
    author="The bioconvert Contributors",
    author_email=metainfo["authors"]["Cokelaer"][1],
    long_description=readme,
    keywords=metainfo["keywords"],
    description=metainfo["description"],
    license=metainfo["license"],
    platforms=metainfo["platforms"],
    url=metainfo["url"],
    download_url=metainfo["download_url"],
    classifiers=metainfo["classifiers"],
    zip_safe=False,
    packages=find_packages(),
    install_requires=requirements,
    extras_require={
        "testing": [
            "pytest",
            "pytest-cov",
            "pytest-env",
            "pytest-xdist",
            "pytest-mock",
            "pytest-timeout",
            "pytest-runner",
            "pytest-rerunfailures",
            "mock",
            "coveralls",
        ],
        "doc": [
            "pillow",
            "sphinx",
            "sphinx_rtd_theme",
            "sphinx_gallery",
            "numpydoc",
            "pygraphviz",
        ],
    },
    # This is recursive include of data files
    exclude_package_data={"": ["__pycache__"]},
    package_data={
        "": ["*.csv", "*.sh"],
        "bioconvert.data": ["*"],
        "bioconvert.misc": ["*"],
    },
    entry_points={
        "console_scripts": [
            "bioconvert=bioconvert.scripts.converter:main",
            "bioconvert_init=bioconvert.scripts.init_convert:main",
            "bioconvert_stats=bioconvert.scripts.stats:main",
        ]
    },
)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Main bioconvert registry that fetches automatically the relevant converter"""
import inspect
import itertools
import pkgutil
import importlib
import colorlog

import bioconvert

_log = colorlog.getLogger(__name__)

__all__ = ["Registry"]


class Registry(object):
    """class to centralise information about available conversions

    ::

        from bioconvert.core.registry import Registry
        r = Registry()
        r.conversion_exists("BAM", "BED")
        r.info()  # returns number of available methods for each converter

        conv_class = r[(".bam", ".bed")]
        converter = conv_class(input_file, output_file)
        converter.convert()

    """

    def __init__(self):
        self._ext_registry = {}
        self._fmt_registry = {}
        self._fill_registry(bioconvert.__path__)
        self._build_path_dict()

    def _fill_registry(
        self, path, target=None, including_not_available_converter=False
    ):
        """
        Explore the directory converters to discover all converter classes
        (a concrete class which inherits from :class:`ConvBase`)
        and fill the register with the input format and output format
        associated to this converter.

        This is called in the constructor once with
        including_not_available_converter set to False and called at any time
        to :meth:`get_all_conversions` with including_not_available_converter
        set to True.

        :param str path: the path of a directory to explore (not recursive)
        :param str target:
        :param bool including_not_available_converter:
        """

        target = self if target is None else target

        def is_converter(item):
            """Check if a module is a converter"""
            obj_name, obj = item
            if not inspect.isclass(obj):
                return False

            # Note that on some Python version, the isabstract is buggy.
            # Therefore, the isabstract does not return False for ConvBase
            # hence the additional check (obj_name in ["ConvBase"])
            return issubclass(obj, bioconvert.ConvBase) and obj_name not in ["ConvBase"]

        modules = pkgutil.iter_modules(path=path)
        for _, module_name, *_ in modules:
            if module_name != "__init__":
                try:
                    module = importlib.import_module("bioconvert." + module_name)
                except (ImportError, TypeError) as err:
                    _log.warning("skip module '{}': {}".format(module_name, err))
                    continue

                converters = inspect.getmembers(module)
                converters = [c for c in converters if is_converter(c)]

                for converter_name, converter in converters:

                    if converter is not None:
                        format_pair = (converter.input_fmt, converter.output_fmt)
                        target[(format_pair)] = converter
                        # have all the combinaisons between the extensions of
                        # output formats of the convertes
                        combo_input_ext = tuple(itertools.product(*converter.input_ext))
                        # have all the combinaisons between the extensions of output
                        # formats of the convertes
                        combo_output_ext = tuple(
                            itertools.product(*converter.output_ext)
                        )
                        all_ext_pair = tuple(
                            itertools.product(combo_input_ext, (combo_output_ext))
                        )
                        for ext_pair in all_ext_pair:
                            if (
                                len(converter.available_methods) == 0
                                and not including_not_available_converter
                            ):
                                _log.debug(
                                    "converter '{}' for {} -> {} was not added as no method is available".format(
                                        converter_name, *ext_pair
                                    )
                                )
                            else:
                                self.set_ext(ext_pair, converter)

    def _build_path_dict(self):
        """
        Construct dictionaries of dictionaries containing shortest paths
        from one format to another.
        """
        from networkx import DiGraph, all_pairs_shortest_path

        # all_pairs_shortest_path yields pairs (n, d) where
        # * n is a node
        # * d is a dict where
        #     * keys are other nodes
        #     * values are shortest paths (i.e. lists of nodes)
        #       from n to these nodes
        # Converting this to dict results in a dict of dicts where
        # * the first key is the source node
        # * the second key is the destination node
        # * the value is a shortest path between these nodes.

        # self._path_dict[in_fmt][out_fmt] = [in_fmt, ..., out_fmt]
        self._path_dict = dict(
            all_pairs_shortest_path(
                # Directed graph of available in_fmt -> out_fmt conversions
                DiGraph(self.get_conversions())
            )
        )

        # self._path_dict_ext[in_ext][out_ext] = [in_ext, ..., out_ext]
        self._path_dict_ext = dict(
            all_pairs_shortest_path(
                # Directed graph of available in_ext -> out_ext conversions
                DiGraph(self.get_conversions_from_ext())
            )
        )

    def conversion_path(self, input_fmt, output_fmt):
        """
        Return a list of conversion steps to get from input and
        output formats

        :param tuple input_fmt:
        :param tuple output_fmt:

        Each step in the list is a pair of formats.
        """
        try:
            fmt_steps = self._path_dict[input_fmt][output_fmt]
        except KeyError:
            fmt_steps = []
        return list(zip(fmt_steps, fmt_steps[1:]))

    def __setitem__(self, format_pair, convertor):
        """
        Register new convertor from input format to output format.

        :param format_pair: the input format, the output format
        :type format_pair: tuple of 2 strings
        :param convertor: the convertor which handle the conversion
                          from input_fmt -> output_fmt
        :type convertor: :class:`ConvBase` object
        """
        if format_pair in self._fmt_registry:
            raise KeyError(
                "an other converter already exists for {} -> {}".format(
                    "_".join(format_pair[0]), "_".join(format_pair[1])
                )
            )
        self._fmt_registry[format_pair] = convertor

    def _check_input_ext(self, ext_pair):
        assert len(ext_pair) == 2, "parameter must be a tuple with 2 items"
        assert isinstance(ext_pair[0], tuple), "first item must be a tuple"
        assert isinstance(ext_pair[1], tuple), "second item must be a tuple"

    def set_ext(self, ext_pair, convertor):
        """
        Register new convertor from input extension and output extension
        in a list. We can have a list of multiple convertors for one
        ext_pair.

        :param tuple ext_pair: tuple containing the input extensions and the
            output extensions e.g. ( ("fastq",) , ("fasta") )
        :param convertor: the convertor which handle the conversion
                          from input_ext -> output_ext
        :type convertor: list of :class:`ConvBase` object
        """
        self._check_input_ext(ext_pair)
        if ext_pair in self._ext_registry:
            self._ext_registry[ext_pair].append(convertor)
        else:
            self._ext_registry[ext_pair] = [convertor]

    def __getitem__(self, format_pair):
        """
        :param format_pair: the input format, the output format
        :type format_pair: tuple of 2 strings
        :return: an object of subclass o :class:`ConvBase`
        """
        format_pair = (format_pair[0], format_pair[1])
        return self._fmt_registry[format_pair]

    def get_ext(self, ext_pair):
        """
        Copy the registry into a dict that behaves like a list
        to be able to have multiple values for a single key
        and from a key have all converter able to do the conversion
        from the input extension to the output extension.

        :param ext_pair: the input extension, the output extension
        :type ext_pair: tuple of 2 strings
        :return: list of objects of subclass o :class:`ConvBase`
        """
        self._check_input_ext(ext_pair)
        return self._ext_registry[ext_pair]

    def __contains__(self, format_pair):
        """
        Can use membership operation on registry to test if a converter
        to go form input format to output format exists.

        :param format_pair: the input format, the output format
        :type format_pair: tuple (or list) of 2 items. The items must be a
            string or a tuple/list of strings.
        :return: True if format_pair is in registry otherwise False.
        """

        # make sure input is tuple of 2 items
        if isinstance(format_pair, (tuple, list)) is False:
            raise ValueError("input argument must be a tuple or list of 2 items")

        if isinstance(format_pair, list):
            format_pair = tuple(format_pair)

        # make sure we have a pair
        if len(format_pair) != 2:
            raise ValueError("input must have 2 items")

        # make sure each item is a string or tuple/list and convert into tuples
        # first item
        if isinstance(format_pair[0], str):
            format_pair = ((format_pair[0],), format_pair[1])
        elif isinstance(format_pair[0], list):
            format_pair = (tuple(format_pair[0]), format_pair[1])

        # second item
        if isinstance(format_pair[1], str):
            format_pair = (format_pair[0], (format_pair[1],))
        elif isinstance(format_pair[1], list):
            format_pair = (format_pair[0], tuple(format_pair[1]))

        # make sure it is upper case
        for item in format_pair[1]:
            item = item.upper()
        for item in format_pair[0]:
            item = item.upper()

        return format_pair in self._fmt_registry

    def __iter__(self):
        """
        Make registry iterable
        through format_pair (str input format, str output format)
        """
        for format_pair in self._fmt_registry:
            yield format_pair

    def get_conversions(self):
        """
        :return: a generator which allow to iterate on all available conversions
                 a conversion is encoded by a tuple of
                 2 strings (input format, output format)
        :retype: generator
        """
        for conv in self._fmt_registry:
            yield conv

    def get_converters_names(self):
        """
        :return: a generator that allows to get the name of the converter
                 from the subclass (ConvBase object)
        :rtype: generator

        """
        for converter in self._fmt_registry.values():
            yield converter.__name__.lower()

    def get_conversions_from_ext(self):
        """
        :return: a generator which allow to iterate on all available conversions
                 a conversion is encoded by a tuple of
                 2 strings (input extension, output extension)
        :rtype: generator
        """
        for conv in self._ext_registry:
            yield conv

    def get_all_conversions(self):
        """
        :return: a generator which allow to iterate on all available
            conversions and their availability;  a conversion is encoded
            by a tuple of 2 strings (input format, output format)
        :retype: generator (input format, output format, status)
        """
        all_converter = {}
        self._fill_registry(
            bioconvert.__path__,
            target=all_converter,
            including_not_available_converter=True,
        )

        for i, o in all_converter:
            yield i, o, (i, o) in self._fmt_registry and len(
                self._fmt_registry[(i, o)].available_methods
            ) > 0

    def conversion_exists(self, input_fmt, output_fmt, allow_indirect=False):
        """
        :param str input_fmt: the input format
        :param str output_fmt: the output format
        :param boolean allow_indirect: whether to count indirect conversions
        :return: True if a converter which transform input_fmt into output_fmt exists
        :rtype: boolean
        """
        input_fmt = tuple([x.upper() for x in input_fmt])
        output_fmt = tuple([x.upper() for x in output_fmt])

        return (input_fmt, output_fmt) in self._fmt_registry or (
            allow_indirect and len(self.conversion_path(input_fmt, output_fmt))
        )

    def get_info(self):
        converters = set([self[this] for this in self._fmt_registry])
        data = {}
        for converter in converters:
            data[converter] = len(converter.available_methods)
        return data

    def iter_converters(self, allow_indirect: bool = False):
        """

        :param bool allow_indirect: also return indirect conversion
        :return: a generator to iterate over (in_fmt, out_fmt, converter
            class when direct, path when indirect)
        :rtype: a generator
        """
        # if allow_indirect:
        for start, stops in self._path_dict.items():
            for stop, path in stops.items():
                if len(path) == 1:
                    pass
                elif len(path) == 2:
                    yield start, stop, self._fmt_registry[(start, stop)], None
                elif allow_indirect:
                    yield start, stop, None, path
        #     return
        # for conv, converter in self._fmt_registry.items():
        #     in_fmt, out_fmt = conv
        #     yield in_fmt, out_fmt, converter, None

    def __str__(self):
        data = self.info()
        C = data["converters"]
        F = data["formats"]
        M = data["methods"]
        txt = "Number of formats: {}".format(F)
        txt += "\n" + "Number of converters: {}".format(C)
        txt += "\n" + "Number of methods : {}".format(M)
        return txt

    def info(self):
        info = self.get_info()
        _converters = [x for x in info.items()]
        _data = [info[k] for k, v in info.items()]
        C = len(_converters)
        M = sum(_data)

        F = len(set([x for items in self.get_all_conversions() for x in items]))

        return {
            "formats": F,
            "converters": C,
            "methods": M,
            "methods_per_converter": round(float(M) / C, 2),
        }



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`BAM` format to :term:`FASTQ` foarmat"""
from bioconvert import ConvBase
from bioconvert.core.decorators import requires
from bioconvert.core.utils import get_extension
import subprocess
import os


class BAM2FASTQ(ConvBase):
    """Convert sorted :term:`BAM` file into :term:`FASTQ` file

    Methods available are based on samtools [SAMTOOLS]_ or bedtools [BEDTOOLS]_.

    .. warning:: Using the bedtools method, the R1 and R2 reads must be next to
        each other so that the reads are sorted similarly

    .. warning:: there is no guarantee that the R1/R2 output file are sorted
        similarly in paired-end case due to supp and second reads

    """

    #: Default value
    _default_method = "samtools"

    def __init__(self, infile, outfile):
        """.. rubric:: constructor

        :param str infile:
        :param str outfile:

        """
        super(BAM2FASTQ, self).__init__(infile, outfile)

    """@requires("bamtools")
    def __method_bamtools(self, *args, **kwargs):

        # this method contains supplementary reads and we don't know
        # what to do with them for now. So, this method is
        # commented. Indeed final R1 and R2 files will not be paired.

        cmd = "bamtools convert -format fastq -in {0} -out {1}".format(
            self.infile, self.outfile
        )
        self.execute(cmd)
    """

    @requires("bedtools")
    def _method_bedtools(self, *args, **kwargs):
        """Do the conversion :term:`BAM` -> :term:`Fastq` using bedtools

        `bedtools documentation <https://bedtools.readthedocs.io/en/latest/>`_"""
        outbasename = os.path.splitext(self.outfile)[0]

        cmd = "bedtools bamtofastq -i {} -fq {}".format(self.infile, self.outfile)
        self.execute(cmd)

        output_ext = get_extension(self.outfile, remove_compression=True)

        # Due to the IO, paired reads are not always consecutive.
        # So, checking the first and second reads for paired data does not
        # work. Instead, we check the first 10 reads and check whether we have
        # at least one paired data.

        data = []
        with open(self.outfile, "r") as fin:
            count = 0
            for i in range(40000):
                x = fin.readline()
                if len(x) == 0:
                    break
                data.append(x)

        from collections import Counter

        isPaired = 2 in Counter([x for i, x in enumerate(data) if i % 4 == 0]).values()
        if isPaired:
            cmd = "bedtools bamtofastq -i {} -fq {}_1.{} -fq2 {}_2.{}".format(
                self.infile, outbasename, output_ext, outbasename, output_ext
            )
            self.execute(cmd)

            # Compress the output if required. We do not use compressor
            # since we may have two outputs.
            comp_ext = get_extension(self.outfile, remove_compression=False)
            if comp_ext in [".gz", ".dsrc", "bz2"]:
                from bioconvert.core.utils import compressor

                compressor("{}_1.{}".format(outbasename, output_ext), comp_ext)
                compressor("{}_2.{}".format(outbasename, output_ext), comp_ext)

    @requires("samtools")
    def _method_samtools(self, *args, **kwargs):
        """Do the conversion :term:`BAM` -> :term:`FASTQ` using samtools

        `SAMtools documentation <http://www.htslib.org/doc/samtools.html>`_"""
        cmd = "samtools fastq {} > {}".format(self.infile, self.outfile)
        self.execute(cmd)
        # Test if input bam file is paired
        p = subprocess.Popen(
            "samtools view -c -f 1 {}".format(self.infile).split(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        isPaired = p.communicate()[0].strip()

        # Collect the extension
        ext = os.path.splitext(self.outfile)[1]

        # FIXME: this compression code may be factorised ?

        output_ext = get_extension(self.outfile, remove_compression=True)

        # If the output file extension is compress extension
        if ext in [".gz", ".bz2", ".dsrc"]:
            outbasename = os.path.splitext(self.outfile)[0].split(".", 1)[0]

            if ext == ".gz":
                compresscmd = "gzip -f"
            elif ext == ".bz2":
                compresscmd = "pbzip2 -f"
            else:
                compresscmd = "dsrc c"

            # When the input file is not paired and the output file needs to be compressed
            if isPaired == "0":
                cmd = "samtools fastq {} > {}.{}".format(
                    self.infile, outbasename, output_ext
                )
                self.execute(cmd)
                if ext == ".dsrc":
                    cmd = "{} {}.{} {}.{}.dsrc".format(
                        compresscmd, outbasename, output_ext, outbasename, output_ext
                    )

                else:
                    cmd = "{} {}.{}".format(compresscmd, outbasename, output_ext)
                self.execute(cmd)
            # When the input file is paired and the output file needs to be compressed
            else:

                cmd = "samtools fastq -1 {}_1.{} -2 {}_2.{} -n {} ".format(
                    outbasename, output_ext, outbasename, output_ext, self.infile
                )
                self.execute(cmd)
                if ext == ".dsrc":
                    cmd = "{} {}_1.{} {}_1.{}.dsrc".format(
                        compresscmd, outbasename, output_ext, outbasename, output_ext
                    )
                    self.execute(cmd)
                    cmd = "{} {}_2.{} {}_2.{}.dsrc".format(
                        compresscmd, outbasename, output_ext, outbasename, output_ext
                    )
                    self.execute(cmd)
                else:
                    cmd = "{} {}_1.{}".format(compresscmd, outbasename, output_ext)
                    self.execute(cmd)
                    cmd = "{} {}_2.{}".format(compresscmd, outbasename, output_ext)
                    self.execute(cmd)
        else:
            outbasename = os.path.splitext(self.outfile)[0]

            # When the input file is not paired
            if isPaired == "0":
                cmd = "samtools fastq {} > {}".format(self.infile, self.outfile)
                self.execute(cmd)
            # When the input file is paired
            else:
                # os.remove(self.outfile)
                cmd = "samtools fastq -1 {}_1.{} -2 {}_2.{} -n {} ".format(
                    outbasename, output_ext, outbasename, output_ext, self.infile
                )
                self.execute(cmd)




cdef char* name = NULL
cdef char* seq = NULL

cdef _fastq2fasta(infile, outfile):


    with open(outfile, "w") as fasta, open(infile, "r") as fastq:
        try:
            while True:
                name, seq, _, _ = next(fastq), next(fastq), next(fastq), next(fastq)
                fasta.write(">{}{}".format(name[1:], seq))
        except StopIteration:
            pass

    """with open(outfile, "w") as fasta, open(infile, "r") as fastq:
       for (name, seq, _) in readfq(fastq):
           fasta.write(">{}\n{}\n".format(name, seq))
    """

def fastq2fasta(infile, outfile):
    _fastq2fasta(infile, outfile)


cdef _fastq2fasta2(infile, outfile):


    with open(outfile, "w") as fasta, open(infile, "r") as fastq:
        try:
            while True:
                name2, seq2, _, _ = next(fastq), next(fastq), next(fastq), next(fastq)
                fasta.write(">{}{}".format(name2[1:], seq2))
        except StopIteration:
            pass

    """with open(outfile, "w") as fasta, open(infile, "r") as fastq:
       for (name, seq, _) in readfq(fastq):
           fasta.write(">{}\n{}\n".format(name, seq))
    """

def fastq2fasta2(infile, outfile):
    _fastq2fasta2(infile, outfile)





###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################

"""Convert :term:`SAM` file to :term:`BAM` format"""
from bioconvert import ConvBase
from bioconvert.core.decorators import requires

import colorlog

logger = colorlog.getLogger(__name__)

__all__ = ["BAM2SAM"]


class BAM2SAM(ConvBase):
    """Convert :term:`BAM` file to :term:`SAM` file


    Methods available are based on samtools [SAMTOOLS]_ , sam-to-bam [SAMTOBAM]_ ,
    sambamba [SAMBAMBA]_ and pysam [PYSAM]_.

    """

    #: default value
    _default_method = "sambamba"
    _threading = True

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile:
        :param str outfile:

        """
        super(BAM2SAM, self).__init__(infile, outfile, *args, **kargs)

    @requires("samtools")
    def _method_samtools(self, *args, **kwargs):
        # -S means ignored (input format is auto-detected)
        # -h means include header in SAM output
        """Here we use the SAMtools tool.

        `SAMtools documentation <http://www.htslib.org/doc/samtools.html>`_"""
        cmd = "samtools view -Sh {} --threads {} -O SAM -o {}"
        cmd = cmd.format(self.infile, self.threads, self.outfile)
        self.execute(cmd)

    @requires(python_library="pysam", external_binary="samtools")
    def _method_pysam(self, *args, **kwargs):
        """We use here the python module Pysam.

        `Pysam documentation <https://pysam.readthedocs.io/en/latest/api.html>`_"""
        import pysam

        pysam.sort("-o", self.outfile, self.infile)

    @requires("sambamba")
    def _method_sambamba(self, *args, **kwargs):
        """Here we use the Sambamba tool.
        This is the default method because it is the fastest.

        `Sambamba documentation <https://lomereiter.github.io/sambamba/docs/sambamba-view.html>`_"""
        cmd = "sambamba view --header {} -o {} -t {}"
        cmd = cmd.format(self.infile, self.outfile, self.threads)
        self.execute(cmd)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018-2019  Institut Pasteur, Paris and CNRS.                #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`WIG` to :term:`BED` format"""

from bioconvert import ConvBase
from bioconvert import requires, compressor

__all__ = ["WIG2BED"]


class WIG2BED(ConvBase):
    """Convert :term:`WIG` file to :term:`BED` file

    Method availabe are based on wig2bed tool.

    """

    #: Default value
    _default_method = "wig2bed"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input WIG file
        :param str outfile: output BED filename

        """
        super(WIG2BED, self).__init__(infile, outfile, *args, **kargs)

    @requires("wig2bed")
    @compressor
    def wig2bed(self, *args, **kwargs):
        """For this method, we use the wig2bed tool.
        
        `wig2bed documentation <https://bedops.readthedocs.io/en/latest/content/reference/file-management/conversion/wig2bed.html>`_"""
        cmd = "wig2bed < {} > {}".format(self.infile, self.outfile)
        self.execute(cmd)
